#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import json
import time
import pytest
from frocket.common.validation.consts import RELATION_OPS, DEFAULT_AGGREGATIONS
from frocket.common.validation.query_validator import QueryValidator
from frocket.common.dataset import DatasetInfo, DatasetId, DatasetColumnType
from frocket.common.dataset import DatasetShortSchema
from frocket.common.validation.error import ValidationErrorKind
from frocket.engine.relation_to_pandas import relation_to_pandas_query
from tests.utils.base_test_utils import SKIP_SLOW_TESTS

ORIGINAL_QUERY_PATH = os.path.dirname(__file__) + '/utils/base_query_example.json'
ORIGINAL_QUERY_JSON = open(ORIGINAL_QUERY_PATH, 'r').read()
ORIGINAL_QUERY_USED_COLUMNS = ['userId', 'ts', 'goalId', 'goalValue', 'eventId', 'eventValue', 'eventType',
                               'device', 'transactionId']
ORIGINAL_QUERY_NAMED_CONDITIONS = {'made_multiple_purchases': 0, 'made_multiple_purchases2': 1, 'seq': 5}
ORIGINAL_QUERY_USED_CONDITIONS = [1, 2, 5, 0]
PANDAS_CONDITION_COL_PREFIX = '__c'  # If the actually-used value changes in the engine, it won't break the tests
ORIGINAL_QUERY_PARSED_RELATION = '(__c1 & __c2) | __c5 | ((__c0))'

TEST_DATASET = DatasetInfo(id=DatasetId.now("testme"), basepath='/dev/null', total_parts=1,
                           timestamp_column='ts', group_id_column='userId')


def query_clone() -> dict:
    return json.loads(ORIGINAL_QUERY_JSON)


def dict_clone(d: dict) -> dict:
    return json.loads(json.dumps(d))


q = query_clone()
TEST_DS_SCHEMA = DatasetShortSchema(
    columns={'userId': DatasetColumnType.INT, 'ts': DatasetColumnType.INT, 'eventId': DatasetColumnType.INT,
             'rowType': DatasetColumnType.STRING, 'eventType': DatasetColumnType.STRING,
             'eventValue': DatasetColumnType.FLOAT, 'goalValue': DatasetColumnType.FLOAT,
             'isCool': DatasetColumnType.BOOL, 'device': DatasetColumnType.STRING,
             'transactionId': DatasetColumnType.INT, 'goalId': DatasetColumnType.INT},
    min_timestamp=q['timeframe']['from'],
    max_timestamp=q['timeframe']['to']
)


def new_validator(query: dict = None, with_dataset: bool = True):
    query = query or query_clone()
    return QueryValidator(source_query=query,
                          dataset=(TEST_DATASET if with_dataset else None),
                          short_schema=(TEST_DS_SCHEMA if with_dataset else None))


def test_original_without_dataset_ok():
    validator = new_validator(with_dataset=False)
    result = validator.expand_and_validate(schema_only=True)
    print(result)
    assert result.success is True
    assert result.error_message is None
    assert result.error_kind is None
    assert result.named_conditions == ORIGINAL_QUERY_NAMED_CONDITIONS
    assert result.used_conditions == ORIGINAL_QUERY_USED_CONDITIONS
    transformed_relation = relation_to_pandas_query(elements=result.relation_elements,
                                                    column_prefix=PANDAS_CONDITION_COL_PREFIX)
    assert transformed_relation == ORIGINAL_QUERY_PARSED_RELATION


def test_used_columns_na():
    validator = new_validator(with_dataset=False)
    result = validator.expand_and_validate(schema_only=True)
    assert result.used_columns is None


def test_dataset_not_passed():
    validator = new_validator(with_dataset=False)
    result = validator.expand_and_validate()
    assert result.error_kind == ValidationErrorKind.INVALID_ARGUMENTS


def test_original_ok():
    validator = new_validator()
    result = validator.expand_and_validate()
    print(result)
    assert result.success is True
    assert result.error_message is None
    assert result.error_kind is None
    assert sorted(result.used_columns) == sorted(ORIGINAL_QUERY_USED_COLUMNS)
    assert result.named_conditions == ORIGINAL_QUERY_NAMED_CONDITIONS
    assert result.used_conditions == ORIGINAL_QUERY_USED_CONDITIONS
    transformed_relation = relation_to_pandas_query(elements=result.relation_elements,
                                                    column_prefix=PANDAS_CONDITION_COL_PREFIX)
    assert transformed_relation == ORIGINAL_QUERY_PARSED_RELATION


def test_original_not_modified():
    orig_query = query_clone()
    passed_query = query_clone()
    validator = new_validator(passed_query)
    result = validator.expand_and_validate()
    assert validator
    assert passed_query == orig_query
    assert passed_query != result.expanded_query


def test_multipass():
    v1 = new_validator()
    res1 = v1.expand_and_validate()
    assert res1.success
    # Now, ensure that feeding the expanded query back into the validator will result in the same expanded query
    # Create a clone to ensure the expanded query dict is not modified when goes into the second validator as source,
    # thus quietly making the test useless...
    expanded1_clone = dict_clone(res1.expanded_query)

    v2 = new_validator(res1.expanded_query)
    res2 = v2.expand_and_validate()
    assert res2.success

    assert res2.expanded_query == expanded1_clone


@pytest.mark.skipif(SKIP_SLOW_TESTS, reason="Skipping slow tests")
def test_many_runs():
    iterations = 50
    max_avg_iteration_time = 0.03
    max_total_time = max_avg_iteration_time * iterations

    q = query_clone()
    start_time = time.time()
    for i in range(iterations):
        res = new_validator(q).expand_and_validate()
        assert res.success

    total_time = time.time() - start_time
    print(f"Time for {iterations} runs: {total_time} seconds")
    assert total_time <= max_total_time


def test_empty():
    v = new_validator({})
    res = v.expand_and_validate()
    assert res.success


def test_additional_property():
    v = new_validator({'hello': 'world'})
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA

    q1 = query_clone()
    q1['query']['conditions'][0]['hello'] = 'world'
    v = new_validator(q1)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA


def test_no_conditions():
    q = {'query': {}}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success and res.used_conditions is None

    q = {'query': {'conditions': []}}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success and res.used_conditions is None


def condition(d: dict, idx: int) -> dict:
    return d['query']['conditions'][idx]


def test_shorthand_filter():
    q = {
        'query': {
            'conditions': [
                {'filter': ['eventId', '==', 1]},
                {'filter': {'column': 'eventId', 'op': '==', 'value': 1}}
            ]
        }
    }
    full_filter = condition(q, 1)['filter']
    assert condition(q, 0)['filter'] != full_filter
    res = new_validator(q).expand_and_validate()
    assert res.success
    assert condition(res.expanded_query, 0)['filter'] == full_filter
    assert condition(res.expanded_query, 1)['filter'] == full_filter


def test_shorthand_and_default_target():
    simple_filter = ['eventId', '==', 1]
    q = {
        'query': {
            'conditions': [
                {'filter': simple_filter},  # No target
                {'filter': simple_filter,  # Shorthand count target
                 'target': ['count', '>=', 1]},
                {'filter': simple_filter,  # Full count target
                 'target': {'type': 'count', 'op': '>=', 'value': 1}},
                {'filter': simple_filter,  # Shorthand sum target
                 'target': ['sum', 'eventValue', '>=', 1]},
                {'filter': simple_filter,  # Full sum target
                 'target': {'type': 'sum', 'op': '>=', 'value': 1, 'column': 'eventValue'}}
            ]
        }
    }
    full_count_target = condition(q, 2)['target']
    full_sum_target = condition(q, 4)['target']
    res = new_validator(q).expand_and_validate()
    assert res.success
    assert condition(res.expanded_query, 0)['target'] == full_count_target
    assert condition(res.expanded_query, 3)['target'] == full_sum_target


def test_valid_condition_names():
    simple_filter = ['eventId', '==', 1]
    valid_names = ['a', 'zrubi_1_a_234', 'PORQUE', 'zz12123123']
    valid_conditions = [{'name': n, 'filter': simple_filter} for n in valid_names]
    q = {'query': {'conditions': valid_conditions}}
    res = new_validator(q).expand_and_validate()
    assert res.success

    invalid_names = ['1', '2asd', '$1', '$adsas', 'import crazy', ' a', 'b '] + RELATION_OPS
    for n in invalid_names:
        q = {'query': {'conditions': [{'name': n, 'filter': simple_filter}]}}
        print(f"Testing query: {q}")
        res = new_validator(q).expand_and_validate()
        assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA


def test_include_zero_validity():
    simple_filter = ['eventId', '==', 1]
    valid_cases = [
        {'target': ['count', '==', 0], 'includeZero': True},
        {'target': ['count', '!=', 1], 'includeZero': True},
        {'target': ['count', '<=', 10], 'includeZero': True},
        {'target': ['count', '<=', 0], 'includeZero': True},
        {'target': ['count', '<', 1], 'includeZero': True},
        {'target': ['count', '>=', 0], 'includeZero': True},
        {'target': ['count', '==', 1], 'includeZero': False},
        {'target': ['count', '!=', 0], 'includeZero': False},
        {'target': ['count', '>', 0], 'includeZero': False},
        {'target': ['count', '>=', 1], 'includeZero': False},
    ]
    invalid_cases = [
        {'target': ['count', '==', 0], 'includeZero': False},
        {'target': ['count', '==', 1], 'includeZero': True},
        {'target': ['count', '!=', 0], 'includeZero': True},
        {'target': ['count', '<', 0], 'includeZero': True},
        {'target': ['count', '>', 0], 'includeZero': True},
        {'target': ['count', '>=', 1], 'includeZero': True},
    ]

    def validate(cases: list, expected_success: bool):
        for case in cases:
            q = {'query': {'conditions': [{**case, 'filter': simple_filter}]}}
            print(f"Testing query: {q}")
            res = new_validator(q).expand_and_validate()
            assert res.success == expected_success
            if not expected_success:
                assert res.error_kind == ValidationErrorKind.TYPE_MISMATCH

    validate(valid_cases, expected_success=True)
    validate(invalid_cases, expected_success=False)

    # Check the one case where 'includeZero' is explicitly expanded if not set, to prevent bugs
    q = {'query': {'conditions': [{'target': ['count', '==', 0],
                                   'filter': simple_filter}]}}
    res = new_validator(q).expand_and_validate()
    assert condition(res.expanded_query, 0)['includeZero']


def test_filters_match_dataset():
    invalid_filters = [
        {'filter': ['eventBlarg', '>', 1]},
        {'filter': ['eventId', 'contains', 'a']},
        {'filter': ['eventId', '>', 'a']},
        {'filter': ['eventId', '==', True]},
        {'filter': ['eventId', '==', False]},
        {'filter': ['rowType', '>', 'a']},
        {'filter': ['rowType', '==', 1]},
        {'filter': ['rowType', '==', True]},
        {'filter': ['isCool', '==', 1]},
        {'filter': ['isCool', '==', 'true']},
        {'filter': ['isCool', 'contains', 'fish']},
        {'filters': [{'column': 'eventId', 'op': '>', 'value': 0},
                     {'column': 'eventBlah', 'op': '>', 'value': 0}]}
    ]

    for case in invalid_filters:
        q = {'query': {'conditions': [case]}}
        print(f"Testing query: {q}")
        res = new_validator(q).expand_and_validate()
        assert not res.success and res.error_kind in [ValidationErrorKind.DATASET_MISMATCH,
                                                      ValidationErrorKind.TYPE_MISMATCH]


def test_targets_match_type_and_dataset():
    simple_filter = ['eventId', '==', 1]
    invalid_targets = [
        ['count', 'contains', 100],
        ['count', '>', True],
        ['count', '>', False],
        ['count', '>', "hello"],
        ['sum', 'jarjar', '>', 100],
        ['sum', 'rowType', '>', 100],
        ['sum', 'eventValue', '>', 'blah'],
        ['sum', 'eventValue', '>', True],
        ['sum', 'eventValue', 'contains', 100],
    ]

    for case in invalid_targets:
        q = {'query': {'conditions': [
            {'filter': simple_filter, 'target': case}
        ]}}
        print(f"Testing query: {q}")
        res = new_validator(q).expand_and_validate()
        assert not res.success and res.error_kind != ValidationErrorKind.UNEXPECTED


def test_name_uniqueness():
    simple_filter = {'column': 'eventId', 'op': '==', 'value': 1}
    q = {
        'query': {
            'conditions': [
                {'name': 'name0', 'filter': simple_filter},
                {'name': 'name1', 'filter': simple_filter},
                {'name': 'name2', 'filter': simple_filter}
            ],
            'aggregations': [
                {"name": "name0", "type": "count", "column": "transactionId"},
                {"name": "name1", "type": "count", "column": "transactionId"},
            ]

        },
        'funnel': {
            'sequence': [{'filter': simple_filter}, {'filter': simple_filter}],
            'stepAggregations': [
                {"name": "name0", "type": "count", "column": "transactionId"},
                {"name": "name1", "type": "count", "column": "transactionId"},
            ],
            'endAggregations': [
                {"name": "name0", "type": "count", "column": "transactionId"},
                {"name": "name1", "type": "count", "column": "transactionId"},
            ]

        }
    }
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success
    assert res.named_conditions == {'name0': 0, 'name1': 1, 'name2': 2}

    invalid_q = dict_clone(q)
    condition(invalid_q, 2)['name'] = 'name0'
    v = new_validator(invalid_q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA

    invalid_q = dict_clone(q)
    invalid_q['query']['aggregations'][1]['name'] = 'name0'
    v = new_validator(invalid_q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA

    invalid_q = dict_clone(q)
    invalid_q['funnel']['stepAggregations'][1]['name'] = 'name0'
    v = new_validator(invalid_q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA

    invalid_q = dict_clone(q)
    invalid_q['funnel']['endAggregations'][1]['name'] = 'name0'
    v = new_validator(invalid_q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA


def test_shorthand_relation():
    simple_filter = {'column': 'eventId', 'op': '==', 'value': 1}
    q = {
        'query': {
            'conditions': [
                {'name': 'name0', 'filter': simple_filter},
                {'name': 'name1', 'filter': simple_filter},
                {'name': 'name2', 'filter': simple_filter}
            ]
        }
    }

    # Expanded missing relation
    v = new_validator(q)
    res = v.expand_and_validate()
    print(res.expanded_query)
    assert res.success and res.expanded_query['query']['relation'] == '$0 and $1 and $2'
    transformed_relation = \
        relation_to_pandas_query(elements=res.relation_elements, column_prefix=PANDAS_CONDITION_COL_PREFIX)
    assert transformed_relation == '__c0 & __c1 & __c2'
    assert res.used_conditions == list(range(3))

    # noinspection PyTypeChecker
    q['query']['relation'] = 'OR'
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success and res.expanded_query['query']['relation'] == '$0 or $1 or $2'
    transformed_relation = \
        relation_to_pandas_query(elements=res.relation_elements, column_prefix=PANDAS_CONDITION_COL_PREFIX)
    assert transformed_relation == '__c0 | __c1 | __c2'
    assert res.used_conditions == list(range(3))

    for relation in ['', '   ', 'mosh']:
        # noinspection PyTypeChecker
        q['query']['relation'] = relation
        v = new_validator(q)
        res = v.expand_and_validate()
        assert res.success is False and res.error_kind == ValidationErrorKind.RELATION


def test_complex_relation():
    simple_filter = {'column': 'eventId', 'op': '==', 'value': 1}
    q = {
        'query': {
            'conditions': [
                {'name': 'name0', 'filter': simple_filter},
                {'name': 'name1', 'filter': simple_filter},
                {'name': 'name2', 'filter': simple_filter}
            ]
        }
    }

    valid_complex_relation = [
        ('$2', '__c2'),

        ('$name2   ', '__c2'),

        ('$0 or $1 and $2 and $name0 or $name2',
         '__c0 | __c1 & __c2 & __c0 | __c2'),

        ('($0 or $1) and  (( $2 and  $name0 ) or (($name2))) ',
         '(__c0 | __c1) & ((__c2 & __c0) | ((__c2)))'),

        (' ( $0  or $1) &&  (( $2 and  $name0)  || ( ($name2 )) ) ',
         '(__c0 | __c1) & ((__c2 & __c0) | ((__c2)))')
    ]

    for source, expected_transformed in valid_complex_relation:
        # noinspection PyTypeChecker
        q['query']['relation'] = source
        print(f"Source relation is '{source}', expected parsed relation: {expected_transformed}")
        v = new_validator(q)
        res = v.expand_and_validate()
        assert res.success
        transformed_source = relation_to_pandas_query(res.relation_elements, column_prefix='__c')
        if transformed_source != expected_transformed:
            raise Exception(f"Expected '{expected_transformed}' but found: '{transformed_source}'")

    invalid_relations = [
        '$1a', '$', '$import() ', '$a-3', '$a a',
        '$3', '$name3', '$97686778343434', '$adlkja(&^%&^$%@#%&^$&*sydfs fs08d6 78s5876596^$%hdjkhasjkdhajkdhjahdjahsj',
        '$1 and', '$1 $2', '$1 and ()', '(($1)))', '()', '$name$1', '$name ($1)', ')', '$1 and(exception)',
        '$1 / $2', '*', '$1 and \'less /etc/gimme', 'sys.exit(1)', '$sys.exit(1)', ';sys.exit(1)', '(sys.exit(1))',
        'raise Exception($1)', '"'
    ]

    for relation in invalid_relations:
        # noinspection PyTypeChecker
        q['query']['relation'] = relation
        print(f"Testing {relation}")
        v = new_validator(q)
        res = v.expand_and_validate()
        assert res.success is False and res.error_kind == ValidationErrorKind.RELATION


def test_aggregation_expansion():
    simple_filter = {'column': 'eventId', 'op': '==', 'value': 1}
    q = {
        'query': {
            'aggregations': [
                {"type": "count", "column": "transactionId"},
                {"type": "sumPerValue", "column": "transactionId", "otherColumn": "eventValue"},
                {"column": "eventId"}
            ]

        },
        'funnel': {
            'sequence': [{'filter': simple_filter}],
            'stepAggregations': [{"column": "eventId"}],
            'endAggregations': [{"column": "eventId"}]

        }
    }

    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success
    original_aggrs = q['query']['aggregations']
    expanded_aggrs = res.expanded_query['query']['aggregations']
    # Aggrs with a type are not changed
    assert expanded_aggrs[0] == original_aggrs[0] and \
           expanded_aggrs[1] == original_aggrs[1]

    # The last aggregation without a type is expanded: types match all DEFAULT_AGGREGATIONS, column name stays
    range_of_expansion = range(len(original_aggrs) - 1, len(expanded_aggrs))
    expanded_types = [expanded_aggrs[i]['type'] for i in range_of_expansion]
    assert sorted(expanded_types) == sorted(DEFAULT_AGGREGATIONS)
    expanded_colnames = [expanded_aggrs[i]['column'] for i in range_of_expansion]
    assert expanded_colnames == [original_aggrs[-1]['column']] * len(DEFAULT_AGGREGATIONS)

    # Expansion happens also in funnel
    assert len(q['funnel']['stepAggregations']) == 1
    assert len(res.expanded_query['funnel']['stepAggregations']) == len(DEFAULT_AGGREGATIONS)
    assert len(q['funnel']['endAggregations']) == 1
    assert len(res.expanded_query['funnel']['endAggregations']) == len(DEFAULT_AGGREGATIONS)


def test_aggregation_other_column():
    simple_filter = {'column': 'eventId', 'op': '==', 'value': 1}
    q = {
        'funnel': {
            'sequence': [{'filter': simple_filter}],
            'endAggregations': [{"column": "eventId", "type": "sumPerValue", "otherColumn": "eventValue"}]

        }
    }

    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success

    qclone = dict_clone(q)
    qclone['funnel']['endAggregations'][0]['otherColumn'] = "what"
    v = new_validator(qclone)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.DATASET_MISMATCH

    qclone['funnel']['endAggregations'][0]['otherColumn'] = "rowType"
    v = new_validator(qclone)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.TYPE_MISMATCH

    qclone['funnel']['endAggregations'][0]['type'] = "countPerValue"
    v = new_validator(qclone)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA

    qclone['funnel']['endAggregations'][0]['type'] = "sumPerValue"
    del qclone['funnel']['endAggregations'][0]['otherColumn']
    v = new_validator(qclone)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA


def test_default_aggr_name():
    simple_filter = {'column': 'eventId', 'op': '==', 'value': 1}
    q = {
        'funnel': {
            'sequence': [{'filter': simple_filter}],
            'stepAggregations': [{"column": "eventId", "type": "count", "name": "aggr1"}, {"column": "eventId"}]

        }
    }

    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success

    # noinspection PyTypeChecker
    q['funnel']['stepAggregations'][1]['name'] = "mename"
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA


def test_timeframe_scale():
    q = query_clone()
    q['timeframe'] = {}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success

    q['timeframe'] = {'from': 123}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.TYPE_MISMATCH

    q['timeframe'] = {'to': 123456789012345}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.TYPE_MISMATCH

    q['timeframe'] = {'from': 1590918400516, 'to': 1600918400516}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success

    q['timeframe'] = {'to': 1590918400516, 'from': 1600918400516}
    v = new_validator(q)
    res = v.expand_and_validate()
    assert res.success is False and res.error_kind == ValidationErrorKind.SCHEMA

# TODO Backlogged tests?
# 1. Sequences: must be filters array, no target, includeZero. Check both for seq conditions and funnel
# 2. Float vs. int columns
# 3. Target 'sum' can accept negative values



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pandas import DataFrame
from frocket.common.tasks.query import QueryResult
from frocket.common.validation.query_validator import QueryValidator
from frocket.engine.query_engine import QueryEngine
import pandas as pd

CATEGORIES = ['fishing', 'running', 'climbing', 'snorkeling']
TYPES = ['view', 'click', 'purchase', 'a2c']
BEGIN_TS = 1610555782
BEGIN_PRICE = 100
GROUP_COLUMN = 'id'
TS_COLUMN = 'timestamp'
ALL_USERS = ['a', 'b', 'c', 'd']
KEYWORDS = ['cat|dog', 'mouse|cat|dog', 'cat', 'dog|mouse|cat']

'''
test cases
==========

target types: V
- sum
- count

numeric operators: V
- <
- >
- ==
- !=
- <=
- >=

string operators: V
- ==
- !=
- contains
- not contains

boolean operators: V
- ==
- !=

aggregations:
- "count"
- "countPerValue"
- "groupsPerValue"
- "sumPerValue"
- "meanPerValue"

'''


def test_sum():
    data = []
    for i in range(1, 1001):
        data.append({
            'f': i,
            'id': 'a',
            'timestamp': i
        })
    data.append({
        'f': 10,
        'id': 'b',
        'timestamp': 2
    })

    data = pd.DataFrame(data)
    query = {
        'conditions': [
            {
                "filter": {
                    "column": "f",
                    "op": ">",
                    "value": 0
                },
                "target": {
                    "type": "sum",
                    "op": "==",
                    "column": "f",
                    "value": int((1000 * 1001) / 2)
                }
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    assert engine_result.query.matching_group_rows == 1000


def test_boolean_operators():
    data = [
        {
            'bool_field': True,
            'id': 'a',
            'timestamp': 1
        },
        {
            'bool_field': False,
            'id': 'b',
            'timestamp': 2
        },
        {
            'id': 'b',
            'timestamp': 3
        }
    ]
    data = pd.DataFrame(data)
    for (op, value, expectation, target_op) in [('==', True, 1, '=='),
                                                # user a has exactly one "true" event, and exactly one row
                                                ('!=', True, 2, '>'),
                                                # user b has two events with value != "True" (one false and one NA)
                                                ('==', False, 2, '=='),
                                                ('!=', False, 3, '==')  # both users have one event that is not False
                                                ]:
        query = {
            'conditions': [
                {
                    "filter": {
                        "column": "bool_field",
                        "op": op,
                        "value": value
                    },
                    "target": {
                        "type": "count",
                        "op": target_op,
                        "value": 1
                    }
                }
            ]
        }
        engine_result = expand_and_run_query(df=data, query_part=query)
        assert engine_result.query.matching_group_rows == expectation


# all queries filter timestamp to be <= BEGIN_TS + 5, which yields 10 rows
# 4 rows for user "a", 3 for "b", 2 for "c" and 1 for "d"
def test_numeric_operators():
    data = test_create_data()
    for (op, val, expectation) in [('==', 2, 1),  # 1 user with 2 events
                                   ('<', 3, 2),  # 2 users, with 1 and 2 events
                                   ('>', 2, 2),  # etc.
                                   ('<=', 1, 1),
                                   ('>=', 4, 1),
                                   ('!=', 1, 3)]:
        query = {
            'relation': 'and',
            'conditions': [
                {
                    "filter": {
                        "column": "timestamp",
                        "op": "<=",
                        "value": BEGIN_TS + 5
                    },
                    "target": {
                        "type": "count",
                        "op": op,
                        "value": val
                    }
                }
            ]
        }

        engine_result = expand_and_run_query(df=data, query_part=query)
        assert engine_result.query.matching_groups == expectation


def expand_and_run_query(df: DataFrame, query_part: dict = None, funnel_part: dict = None) -> QueryResult:
    full_query = {}
    if query_part:
        full_query['query'] = query_part
    if funnel_part:
        full_query['funnel'] = funnel_part
    validator = QueryValidator(full_query)
    print(f"Query before expansion: {full_query}")
    validation_result = validator.expand_and_validate(schema_only=True)
    expanded_query = validation_result.expanded_query
    print(f"Query after expansion: {expanded_query}")
    print(validation_result.error_message)
    assert validation_result.success
    engine = QueryEngine(group_by_column=GROUP_COLUMN, timestamp_column=TS_COLUMN)
    return engine.run(df, expanded_query)


def test_create_data():
    # 4 users
    dfs = []
    for index, user in enumerate(ALL_USERS):
        index += 2
        number_of_rows_for_user = 2 ** index
        data = {
            'id': [user] * number_of_rows_for_user,
            'timestamp': [BEGIN_TS + index + i for i in range(number_of_rows_for_user)],
            'category': CATEGORIES * int(number_of_rows_for_user / 4),
            'type': TYPES * int(number_of_rows_for_user / 4),
            'price': [BEGIN_PRICE + index + i for i in range(number_of_rows_for_user)],
            'keywords': [f'{w}_{user}' for w in KEYWORDS] * int(number_of_rows_for_user / 4)
        }
        dfs.append(pd.DataFrame.from_dict(data))
    return pd.concat(dfs, axis=0)


def test_empty_query_brings_all_users():
    data = test_create_data()
    query = {
        'conditions': []
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    assert engine_result.query.matching_groups == len(ALL_USERS)
    assert engine_result.query.matching_group_rows == len(data)


def test_simple_filter():
    data = test_create_data()
    query = {
        'relation': 'and',
        'conditions': [
            {
                "filter": {
                    "column": "price",
                    "op": ">",
                    "value": 133
                },
                "target": {
                    "type": "count",
                    "op": ">",
                    "value": 2
                }
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    # this matches only user 'd', that has 2^5 rows
    assert engine_result.query.matching_groups == 1
    assert engine_result.query.matching_group_rows == 2 ** 5


def test_logical_or():
    data = test_create_data()
    query = {
        'relation': 'or',
        'conditions': [
            {
                "filter": {
                    "column": "price",
                    "op": "<=",
                    "value": 104
                },
                "target": {
                    "type": "count",
                    "op": ">",
                    "value": 2
                }
            },
            {
                "filter": {
                    "column": "price",
                    "op": ">",
                    "value": 133
                },
                "target": {
                    "type": "count",
                    "op": ">",
                    "value": 2
                }
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    # first condition matches user 'a', that has 2^2 rows,
    # logical or with condition from 'simple' test adds user 'd'
    assert engine_result.query.matching_groups == 1 + 1
    assert engine_result.query.matching_group_rows == (2 ** 2) + (2 ** 5)


def test_column_aggregations():
    data = test_create_data()
    query = {
        'conditions': [],
        'aggregations': [
            {
                'column': 'category'
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    aggs = engine_result.query.aggregations
    assert len(aggs) == 3
    aggs_d = {agg.type: agg for agg in aggs}
    assert aggs_d['count'].value == len(data)
    assert aggs_d['countPerValue'].value == {cat: 15 for cat in CATEGORIES}
    assert aggs_d['groupsPerValue'].value == {cat: 4 for cat in CATEGORIES}


def test_other_column_aggregations():
    data = []
    for i in range(10):
        d = {
            'id': 'a',
            'timestamp': 1,
            'category': 'a' if i % 2 == 0 else 'b',
            'price': i
        }
        data.append(d)
    data = pd.DataFrame(data)
    query = {
        'conditions': [],
        'aggregations': [
            {
                'column': 'category',
                'type': 'sumPerValue',
                'otherColumn': 'price'
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    aggs = engine_result.query.aggregations
    '''
cat  price
a      0
b      1
a      2
b      3
a      4
b      5
a      6
b      7
a      8
b      9
    '''
    assert aggs[0].value['a'] == 0 + 2 + 4 + 6 + 8
    assert aggs[0].value['b'] == 1 + 3 + 5 + 7 + 9
    query = {
        'conditions': [],
        'aggregations': [
            {
                'column': 'category',
                'type': 'meanPerValue',
                'otherColumn': 'price'
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    aggs = engine_result.query.aggregations
    assert aggs[0].value['a'] == (0 + 2 + 4 + 6 + 8) / 5
    assert aggs[0].value['b'] == (1 + 3 + 5 + 7 + 9) / 5


def test_multi_filter_simple():
    data = test_create_data()
    query = {
        'relation': 'or',
        'conditions': [
            {
                "filters": [
                    {
                        "column": "price",
                        "op": "<=",
                        "value": 104
                    },
                    {
                        "column": "category",
                        "op": "==",
                        "value": "running"
                    }
                ],
                "target": {
                    "type": "count",
                    "op": "==",
                    "value": 1
                }
            }
        ]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    # this query matches users a and b, which have 2^2 + 2^3 rows together
    assert engine_result.query.matching_groups == 1 + 1
    assert engine_result.query.matching_group_rows == (2 ** 2) + (2 ** 3)


def test_string_operators():
    data = test_create_data()
    print(len(data))
    # all conditions are with > 3 count target
    for (op, value, expectation) in [('contains', 'dog_a', 0),  # a has 3 events containing dog_a
                                     ('contains', 'dog_b', 8),  # b has 8 etc.
                                     ('contains', 'dog_c', 16),
                                     ('==', 'cat_a', 0),  # same as above, only 1 cat_a event
                                     ('==', 'cat_c', 16),
                                     ('!=', 'cat_a', 56),  # all but user a have more than 3 !cat_a events
                                     ('not contains', 'mouse', 56),  # user "a" has only 2 events that doesn't
                                     # contain "mouse", and he has overall 4 events. All other users have more than 3
                                     # events that doesn't contain "mouse"
                                     ]:
        query = {
            'relation': 'or',
            'conditions': [
                {
                    "filter": {
                        "column": "keywords",
                        "op": op,
                        "value": value
                    },
                    "target": {
                        "type": "count",
                        "op": ">",
                        "value": 3
                    }
                }
            ]
        }
        engine_result = expand_and_run_query(df=data, query_part=query)
        assert engine_result.query.matching_group_rows == expectation

def test_sequence_condition():
    data = []
    for i in range(5):
        d = [
            # user 'a' will have this sequence of categories: 0 1 2 3 4
            {
            'id': 'a',
            'timestamp': i,
            'category': i,
            'price': 1
        },
            # user 'b' will have this sequence of categories: 0 1 -1 -1 -1
        {
            'id': 'b',
            'timestamp': i,
            'category': i if i < 2 else -1,
            'price': 2
        },
            # user 'c' will have this sequence of categories: 4 3 2 1 0
        {
            'id': 'c',
            'timestamp': i,
            'category': 4 - i,
            'price': 3
        }
        ]
        data.extend(d)
    data = pd.DataFrame(data)
    query = {
        'conditions': [
            {
                "sequence": [
                    {
                        "filter": ["category", "==", 0]
                    },
                    {
                        "filter": ["category", "<=", 2]
                    },
                    {
                        "filter": ["category", "<", 0]
                    }
                ]
            }
        ],
        'aggregations': [{'column': 'price'}]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    # only user b fits the query
    assert engine_result.query.matching_groups == 1
    assert '2' in engine_result.query.aggregations[-1].value
    query = {
        'conditions': [
            {
                "sequence": [
                    {
                        "filter": ["category", ">=", 3]
                    },
                    {
                        "filter": ["category", "<=", 2]
                    },
                    {
                        "filter": ["category", "==", 0]
                    }
                ]
            }
        ],
        'aggregations': [{ 'column': 'price'}]
    }
    engine_result = expand_and_run_query(df=data, query_part=query)
    # this time it's user c
    assert engine_result.query.matching_groups == 1
    assert '3' in engine_result.query.aggregations[-1].value




#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from frocket.common.config import config
from frocket.common.helpers.utils import memoize
from frocket.datastore.datastore import Datastore
from frocket.datastore.blobstore import Blobstore
from frocket.datastore.redis_store import RedisStore

logger = logging.getLogger(__name__)

DATASTORE_CLASSES = {
    "redis": RedisStore,
}

BLOBSTORE_CLASSES = {
    "redis": RedisStore,
}


# TODO backlog consider thread-safety here: while RedisStore is thread-safe and having more than one is ok, future
#  implementations may not be? (or should be required to)
def _get_store(store_kind: str, store_mapping: dict):
    store_class = store_mapping[config.get(store_kind).lower()]
    store = store_class(role=store_kind)
    logger.info(f"Initialized {store}")
    return store


@memoize
def get_datastore() -> Datastore:
    return _get_store("datastore", DATASTORE_CLASSES)


@memoize
def get_blobstore() -> Blobstore:
    return _get_store("blobstore", BLOBSTORE_CLASSES)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import boto3
from frocket.common.config import config, ConfigDict
from frocket.common.helpers.utils import timestamped_uuid, memoize

SKIP_S3_TESTS = os.environ.get('SKIP_S3_TESTS', "False").lower() == 'true'


@memoize
def _init_mock_s3_config():
    if SKIP_S3_TESTS:
        print(f"Skipping mock S3 config")
    config['s3.aws.endpoint.url'] = \
        os.environ.get('MOCK_S3_URL', config.get('s3.aws.endpoint.url', 'http://localhost:9000'))
    config['s3.aws.access.key.id'] = \
        os.environ.get('MOCK_S3_USER', config.get('s3.aws.access.key.id', 'testonly'))
    config['s3.aws.secret.access.key'] = \
        os.environ.get('MOCK_S3_SERCET', config.get('s3.aws.secret.access.key', 'testonly'))


def mock_s3_env_variables():
    _init_mock_s3_config()
    return {
        ConfigDict.to_env_variable(key): config.get(key)
        for key in ['s3.aws.endpoint.url', 's3.aws.access.key.id', 's3.aws.secret.access.key']
    }


def new_mock_s3_bucket():
    if SKIP_S3_TESTS:
        return None
    _init_mock_s3_config()

    bucket_name = timestamped_uuid('testbucket-')
    s3 = boto3.resource('s3', **config.aws_client_settings(service='s3'))
    bucket = s3.Bucket(bucket_name)
    bucket.create()
    print(f"Bucket '{bucket_name}' created")
    return bucket



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os
import time
from dataclasses import dataclass
from enum import auto
from typing import Optional, Union
import pytest
import requests
from frocket.common.helpers.utils import timestamped_uuid
from frocket.common.serializable import AutoNamedEnum
from tests.utils.dataset_utils import new_test_dataset, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, TIME_SHIFT, \
    BASE_TIME, TestColumn, BASE_USER_ID, DEFAULT_GROUP_COUNT, datafile_schema

# TODO as part of automated failure/retry tests: test query timeout
from tests.utils.mock_s3_utils import SKIP_S3_TESTS

TEST_API_SERVER_URL = os.environ.get('TEST_API_SERVER_URL', 'http://127.0.0.1:5000/')
if not TEST_API_SERVER_URL.endswith('/'):
    TEST_API_SERVER_URL += '/'
ERROR_STRING_PREFIX = 'Error: '
LIST_DATASETS_EXPECTED_KEYS = sorted(['basepath', 'groupIdColumn', 'id', 'timestampColumn', 'totalParts'])
STREAM_STATUS_EXPECTES_KEYS = ['message', 'stage', 'tasks']


class UrlMethod(AutoNamedEnum):
    GET = auto()
    POST = auto()


@dataclass
class UrlInfo:
    url: str
    method: UrlMethod
    returns_success_key: bool = True


@dataclass
class ResponseInfo:
    text: str
    json: Optional[Union[dict, list]]


class URLs:
    @staticmethod
    def register_url(run_async: bool = False) -> UrlInfo:
        return UrlInfo(
            url=f"{TEST_API_SERVER_URL}datasets/register?stream={str(run_async).lower()}",
            method=UrlMethod.POST)

    @staticmethod
    def unregister_url(ds_name: str, force: bool = False) -> UrlInfo:
        return UrlInfo(
            url=f"{TEST_API_SERVER_URL}datasets/{ds_name}/unregister?force={str(force).lower()}",
            method=UrlMethod.POST)

    @staticmethod
    def list_datasets_url() -> UrlInfo:
        return UrlInfo(url=f"{TEST_API_SERVER_URL}datasets", method=UrlMethod.GET,
                       returns_success_key=False)

    @staticmethod
    def dataset_parts_url(ds_name: str) -> UrlInfo:
        return UrlInfo(url=f"{TEST_API_SERVER_URL}datasets/{ds_name}/parts", method=UrlMethod.GET,
                       returns_success_key=False)

    @staticmethod
    def dataset_schema_url(ds_name: str, full: bool = True) -> UrlInfo:
        return UrlInfo(
            url=f"{TEST_API_SERVER_URL}datasets/{ds_name}/schema?full={str(full).lower()}",
            method=UrlMethod.GET, returns_success_key=False)

    @staticmethod
    def query_url(ds_name: str, run_async: bool = False) -> UrlInfo:
        return UrlInfo(
            url=f"{TEST_API_SERVER_URL}datasets/{ds_name}/query?stream={str(run_async).lower()}",
            method=UrlMethod.POST)

    @staticmethod
    def run(url_info: UrlInfo,
            json_body: dict = None,
            expected_success: bool = True,
            expect_stream: bool = False,
            errors_are_json: bool = True,
            expected_dict: dict = None,
            stream_expected_task_count: int = None) -> ResponseInfo:
        # print(f"UrlInfo: {url_info}, json_body: {json_body}, expected success: {expected_success}, "
        #       f"expected dict: {expected_dict}")
        if url_info.method == UrlMethod.POST:
            response = requests.post(url_info.url, json=json_body, stream=expect_stream)
        elif url_info.method == UrlMethod.GET:
            response = requests.get(url_info.url, stream=expect_stream)
        else:
            raise Exception('Uknown method!')

        if expect_stream:
            lines = [line.decode() for line in response.iter_lines() if line]
            if len(lines) > 1:
                for status_line in lines[:-1]:
                    status = json.loads(status_line)
                    assert sorted(status.keys()) == STREAM_STATUS_EXPECTES_KEYS
                    if stream_expected_task_count and status['tasks']:
                        total_tasks = sum(status['tasks'].values())
                        assert total_tasks == stream_expected_task_count
            response_text = lines[-1]
        else:
            response_text = response.text

        if expected_success or errors_are_json:
            response_json = json.loads(response_text)
        else:
            response_json = None

        # print(f"Response as text: {response_text}")
        if expected_success:
            assert response.status_code == 200
            assert type(response_json) in [dict, list]
            if url_info.returns_success_key:
                assert response_json['success'] is True
                assert not response_json.get('errorMessage', None)
        else:
            print(f"Response as text: {response_text}")
            assert response.status_code == 500 if not expect_stream else 200
            if errors_are_json:
                if url_info.returns_success_key:
                    assert response_json['success'] is False
                    assert response_json['errorMessage'] is not None
            else:
                assert response_text.startswith(ERROR_STRING_PREFIX)

        if expected_dict:
            for k, v in expected_dict.items():
                assert response_json[k] == v

        return ResponseInfo(response_text, response_json)


def build_register_args(basepath: str, ds_name: str = None):
    ds_name = ds_name or timestamped_uuid(prefix="ds-")
    return {
        "name": ds_name,
        "basepath": basepath,
        "group_id_column": DEFAULT_GROUP_COLUMN,
        "timestamp_column": DEFAULT_TIMESTAMP_COLUMN,
        "validation_mode": "FIRST_LAST",
        "validate_uniques": True
    }


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_register_unregister_ok():
    with new_test_dataset(4) as test_ds:
        args = build_register_args(basepath=test_ds.copy_to_s3())
        ds_name = args['name']
        URLs.run(url_info=URLs.register_url(run_async=False), json_body=args)
        URLs.run(url_info=URLs.unregister_url(ds_name=ds_name), expected_dict={'datasetFound': True})
        URLs.run(url_info=URLs.unregister_url(ds_name=ds_name), expected_dict={'datasetFound': False})


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_register_invalid_path():
    args = build_register_args(basepath="s3://thisisnot/apath/")
    URLs.run(url_info=URLs.register_url(run_async=False), json_body=args,
             expected_success=False, errors_are_json=True)


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_register_invalid_column():
    with new_test_dataset(1) as test_ds:
        args = build_register_args(basepath=test_ds.copy_to_s3())
        args['group_id_column'] = 'notagroupidcolumn'
        URLs.run(url_info=URLs.register_url(run_async=False), json_body=args,
                 expected_success=False, errors_are_json=True)


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_register_invalid_args():
    with new_test_dataset(1) as test_ds:
        s3path = test_ds.copy_to_s3()
        args = build_register_args(s3path)
        args['validation_mode'] = 'SINGLEX'  # Invalid
        URLs.run(url_info=URLs.register_url(run_async=False), json_body=args,
                 expected_success=False, errors_are_json=False)

        args = build_register_args(s3path)
        args['validate_uniques'] = '123'  # Invalid
        URLs.run(url_info=URLs.register_url(run_async=False), json_body=args,
                 expected_success=False, errors_are_json=False)


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_register_async():
    with new_test_dataset(2) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        URLs.run(url_info=URLs.register_url(run_async=True), json_body=args,
                 expect_stream=True, stream_expected_task_count=2)
        URLs.run(url_info=URLs.unregister_url(ds_name=args['name']), expected_dict={'datasetFound': True})


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_list_datasets():
    ds_list = URLs.run(url_info=URLs.list_datasets_url()).json
    assert type(ds_list) is list
    for e in ds_list:
        assert sorted(e.keys()) == LIST_DATASETS_EXPECTED_KEYS
    original_count = len(ds_list)

    num_parts = 4
    with new_test_dataset(num_parts) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        ds_name = args['name']
        URLs.run(url_info=URLs.register_url(), json_body=args)

    updated_list = URLs.run(url_info=URLs.list_datasets_url()).json
    assert len(updated_list) == original_count + 1
    found = [e for e in updated_list if e['id']['name'] == ds_name]
    assert len(found) == 1
    assert found[0]['basepath'] == args['basepath'] and found[0]['totalParts'] == num_parts

    URLs.run(URLs.unregister_url(ds_name))
    updated_list = URLs.run(url_info=URLs.list_datasets_url()).json
    assert len(updated_list) == original_count


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_unregister_force():
    with new_test_dataset(1) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        ds_name = args['name']
        URLs.run(url_info=URLs.register_url(), json_body=args)
        curr_second = int(time.time())
        URLs.run(url_info=URLs.query_url(ds_name), json_body={})

        response = URLs.run(url_info=URLs.unregister_url(ds_name=args['name'], force=False),
                            expected_success=False, errors_are_json=True)
        assert curr_second <= response.json['datasetLastUsed'] < curr_second + 10
        assert 'less than safety interval' in response.json['errorMessage']

        URLs.run(url_info=URLs.unregister_url(ds_name=args['name'], force=True),
                 expected_success=True, expected_dict={'datasetFound': True})
        URLs.run(url_info=URLs.unregister_url(ds_name=args['name'], force=False),
                 expected_success=True, expected_dict={'datasetFound': False})


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_query_ok():
    num_parts = 3
    with new_test_dataset(num_parts) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        ds_name = args['name']
        URLs.run(url_info=URLs.register_url(), json_body=args)

        query = {
            'query': {
                'conditions': [
                    # Should exclude exactly one user (one with lowest ID)
                    {'filter': [DEFAULT_GROUP_COLUMN, '>', BASE_USER_ID]}
                ]
            },
            'funnel': {
                'sequence': [
                    # Should include exactly one user (one with higehst ID)
                    {'filter': [DEFAULT_TIMESTAMP_COLUMN, '>=', BASE_TIME + (TIME_SHIFT * num_parts)]}
                ],
                "endAggregations": [
                    {'column': TestColumn.bool, 'type': 'countPerValue'}
                ]
            }
        }
        response = URLs.run(url_info=URLs.query_url(ds_name), json_body=query)
        assert response.json['query']['matchingGroups'] == (DEFAULT_GROUP_COUNT * 2) - 1
        assert response.json['funnel']['sequence'][0]['matchingGroups'] == 1
        assert sum(response.json['funnel']['endAggregations'][0]['value'].values()) == \
               response.json['funnel']['sequence'][0]['matchingGroupRows']

        URLs.run(url_info=URLs.unregister_url(ds_name=args['name'], force=True),
                 expected_success=True, expected_dict={'datasetFound': True})


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_query_invalid_args():
    """Errors that should be raise already on invoker side, and result in 'Error: ...' string response"""
    with new_test_dataset(1) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        ds_name = args['name']
        URLs.run(url_info=URLs.register_url(), json_body=args)

        response = URLs.run(url_info=URLs.query_url(ds_name), json_body={'blah': 123},
                            expected_success=False, errors_are_json=False)
        assert "'blah' was unexpected" in response.text

        invalid_ds_name = ds_name + 'blah!!!'
        response = URLs.run(url_info=URLs.query_url(invalid_ds_name), json_body={},
                            expected_success=False, errors_are_json=False)
        assert f"Dataset {invalid_ds_name} not found" in response.text

        URLs.run(url_info=URLs.unregister_url(ds_name, force=True))


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_query_missing_files():
    """Missing files (after registration) should cause workers to fail"""
    with new_test_dataset(2) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        args['validation_mode'] = 'SINGLE'
        ds_name = args['name']

        # Query works, initially
        URLs.run(url_info=URLs.register_url(), json_body=args)
        URLs.run(url_info=URLs.query_url(ds_name), json_body={})

        # Now re-register; files already cached should be disregarded,
        # and register itself is set to only validate a SINGLE file (that would be cached).
        # Meaning, other parts should not be locally cached, and loading from source should now fail
        URLs.run(url_info=URLs.register_url(), json_body=args)
        for key in test_ds.bucket.objects.all():
            key.delete()

        # Expected to fail "as JSON" (with a status and errorMessage), as it's errors from workers
        response = URLs.run(url_info=URLs.query_url(ds_name), json_body={},
                            expected_success=False, errors_are_json=True)

        URLs.run(url_info=URLs.unregister_url(ds_name, force=True))


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_dataset_details():
    """A non-exhaustive test of fetching dataset parts and schema info (this was tested in detail elsewhere)"""
    with new_test_dataset(2) as test_ds:
        args = build_register_args(test_ds.copy_to_s3())
        ds_name = args['name']
        URLs.run(url_info=URLs.register_url(), json_body=args)

        ds_parts = URLs.run(url_info=URLs.dataset_parts_url(ds_name)).json
        assert ds_parts['filenames'] == test_ds.expected_parts.filenames
        expected_columns = json.loads(datafile_schema().to_json())['columns']

        ds_short_schema = URLs.run(url_info=URLs.dataset_schema_url(ds_name, full=False)).json
        assert ds_short_schema['columns'] == expected_columns

        ds_full_schema = URLs.run(url_info=URLs.dataset_schema_url(ds_name, full=True)).json
        assert ds_full_schema['columns'][DEFAULT_TIMESTAMP_COLUMN]['colattrs']['numericMin'] == BASE_TIME

        URLs.run(url_info=URLs.unregister_url(ds_name))



"""
Consts and types for the query validation package
TODO backlog create a nice enum for all query keywords
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os
import re
from pathlib import Path
from typing import Dict, NamedTuple
from frocket.common.dataset import DatasetColumnType
from frocket.common.validation.path_visitor import PathVisitor

# JSON Schema file
QUERY_SCHEMA_LOCATION = Path(os.path.dirname(__file__)) / '../../resources/query_schema.json'
QUERY_SCHEMA = json.load(open(QUERY_SCHEMA_LOCATION, 'r'))

TARGET_TYPES_WITH_INCLUDE_ZERO = ['count']
TARGET_OPS_SUPPORTING_INCLUDE_ZERO = ['<', '<=', '==', '!=', '>=']
TARGET_TYPES_WITH_OTHER_COLUMN = ['sum']
AGGR_TYPES_WITH_OTHER_COLUMN = ['sumPerValue', 'meanPerValue']
DEFAULT_TARGET = {'type': 'count', 'op': '>=', 'value': 1}
DEFAULT_AGGREGATIONS = ['count', 'countPerValue', 'groupsPerValue']
AGGREGATIONS_PATHS = ['query.aggregations',
                      'funnel.stepAggregations',
                      'funnel.endAggregations']
SINGLE_FILTER_PATHS = ['query.conditions.filter',
                       'query.conditions.sequence.filter',
                       'funnel.sequence.filter']
FILTER_ARRAY_PATHS = ['query.conditions.filters',
                      'query.conditions.sequence.filters',
                      'funnel.sequence.filters']

VALID_IDENTIFIER_PATTERN = re.compile(r'[A-Z][A-Z_0-9]*$', re.IGNORECASE)
UNIQUE_IDENTIFIER_SCOPES = ['query.conditions.name'] + \
                           [f"{path}.name" for path in AGGREGATIONS_PATHS]

EQUALITY_OPERATORS = ['==', '!=']
NUMERIC_OPERATORS = [*EQUALITY_OPERATORS, '>', '>=', '<', '<=']
STRING_OPERATORS = [*EQUALITY_OPERATORS, 'contains', 'regex']
OPERATORS_BY_COLTYPE = {
    DatasetColumnType.INT: NUMERIC_OPERATORS,
    DatasetColumnType.FLOAT: NUMERIC_OPERATORS,
    DatasetColumnType.BOOL: EQUALITY_OPERATORS,
    DatasetColumnType.STRING: STRING_OPERATORS
}
VALUE_TYPES_BY_COLTYPE = {
    DatasetColumnType.INT: [int],
    DatasetColumnType.FLOAT: [int, float],
    DatasetColumnType.BOOL: [bool],
    DatasetColumnType.STRING: [str]
}
NUMERIC_COLTYPES = [DatasetColumnType.INT, DatasetColumnType.FLOAT]

RELATION_OPS = ['and', 'or', '||', '&&']
DEFAULT_RELATION_OP = 'and'
CONDITION_COLUMN_PREFIX = "__cond_"


class QueryConditionsMap(NamedTuple):
    count: int
    names: Dict[str, int]


def map_condition_names(query: dict) -> QueryConditionsMap:
    """Map named conditions (which is optional) to the condition ID (index in conditions list)."""
    conditions = PathVisitor(query, 'query.conditions').list()
    names = {cond['name'].strip().lower(): i
             for i, cond in enumerate(conditions) if 'name' in cond}
    return QueryConditionsMap(count=len(conditions), names=names)



"""
Task request/response classes for the registration job (discovering, validating and storing metadata for a dataset)
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from enum import auto
from typing import Optional
from frocket.common.dataset import DatasetInfo, DatasetPartId, DatasetSchema
from frocket.common.serializable import SerializableDataClass, AutoNamedEnum, enveloped
from frocket.common.tasks.base import BaseTaskRequest, BaseTaskResult, BlobId, BaseJobResult, BaseApiResult


class DatasetValidationMode(AutoNamedEnum):
    SINGLE = auto()  # Only validate a single file in the dataset (meaning no cross-file consistency checks are done!)
    FIRST_LAST = auto()  # Validate only first and last files (by lexicographic sorting) and cross-check them
    SAMPLE = auto()  # Takes a sample of files, proportional to the no.o of files and up to a configured maximum.


REGISTER_DEFAULT_FILENAME_PATTERN = '*.parquet'  # Ignore files such as '_SUCCESS' and the like in discovery
REGISTER_DEFAULT_VALIDATION_MODE = DatasetValidationMode.SAMPLE
REGISTER_DEFAULT_VALIDATE_UNIQUES = True


@dataclass(frozen=True)
class RegisterArgs(SerializableDataClass):
    """Parameters collected by the CLI / API server for the registration job"""
    name: str
    basepath: str
    group_id_column: str
    timestamp_column: str
    pattern: str = REGISTER_DEFAULT_FILENAME_PATTERN
    validation_mode: DatasetValidationMode = REGISTER_DEFAULT_VALIDATION_MODE
    validate_uniques: bool = REGISTER_DEFAULT_VALIDATE_UNIQUES


@enveloped
@dataclass(frozen=True)
class RegistrationTaskRequest(BaseTaskRequest):
    dataset: DatasetInfo
    part_id: DatasetPartId
    # If RegisterArgs.validate_uniques=true, task should return all group IDs in file
    return_group_ids: bool


@enveloped
@dataclass(frozen=True)
class RegistrationTaskResult(BaseTaskResult):
    dataset_schema: Optional[DatasetSchema]  # None on failures
    part_id: DatasetPartId
    # If RegistrationTaskRequest.return_group_ids=true, a reference to the blob with the group IDs
    group_ids_blob_id: Optional[BlobId]


@dataclass(frozen=True)
class RegistrationJobResult(BaseJobResult):
    dataset: DatasetInfo


@dataclass(frozen=True)
class UnregisterApiResult(BaseApiResult):
    dataset_found: bool
    dataset_last_used: Optional[float]



"""
This is the most generic implementation for getting runtime-environment metrics:
it does not assume we know the cost of the host machine for the request duration,
and getting physical memory size should generally work on Linux variants and OS X versions.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
from frocket.common.metrics import EnvironmentMetricsProvider, MetricData, MetricName

logger = logging.getLogger(__name__)


class GenericEnvMetricsProvider(EnvironmentMetricsProvider):
    def _memory_bytes(self):
        # Tested on Linux and OS X
        try:
            mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
        except ValueError:
            # Fallback to sysctl in case that os.sysconf('SC_PHYS_PAGES') fails on OS X (seems version specific)
            # noinspection PyBroadException
            try:
                stream = os.popen('sysctl hw.memsize')
                mem_bytes = int(stream.read().split(' ')[1])
            except Exception as e:
                logger.warning(f"Can't detect machine memory: {e}")
                return None

        return MetricData(MetricName.MACHINE_MEMORY_BYTES, mem_bytes)

    def _cost_dollars(self, duration=None):
        return None



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, Any, Optional

PathVisitorCallback = Callable[[Any], Optional[Any]]


class PathVisitor:
    """
    A helper class for safely fetching nested attributes in a dictionary.
    It is used extensively by the QueryValidator to extract and transform nested attributes.

    The class is instantiated with a root dict and a dot-delimited string path (e.g. 'attr.sub_attr.sub_sub').
    Then, visit() can be called once (or more) to run code over the matching value/s, if any. If the key is not found,
    no error is thrown. list() is a convenience method which visits the elements and returns them as a list,
    returning an empty list on no matches.

    By default, if the leaf key is a list, the visitor function is called for each element.
    However, if the list itself is what you need, pass list_to_items=False on init.

    Modifying attributes *below* the visited value is safe (be it a dict, a list, an object), however sometimes you
    may want to replace the whole value itself being itereated. For example, the QueryValidator replaces shorthand-
    notation objects, which are lists, into full-notation dicts.
    To support that, init the object with modifiable=true and return the replacement value from the visitor function,
    or None to keep the value.

    For usage examples, see test_path_visitor.py.
    """
    _KEY_NOT_FOUND = object()

    def __init__(self, root: dict, path: str, modifiable: bool = False, list_to_items: bool = True):
        assert (isinstance(root, dict))
        self._root = root
        self._paths = path.strip().split(".")
        self._modifiable = modifiable
        self._list_to_items = list_to_items

    def visit(self, func: PathVisitorCallback):
        if len(self._paths) > 0:
            self._visit_dict(self._root, 0, func)

    def list(self) -> list:
        result = []
        self.visit(lambda v: result.append(v))
        return result

    def _visit_dict(self, d: dict, depth: int, func: PathVisitorCallback):
        v = d.get(self._paths[depth], self._KEY_NOT_FOUND)  # Differentiate a None value from an inexisting key
        if v == self._KEY_NOT_FOUND:
            return  # Bumped into a wall

        if isinstance(v, list) and self._list_to_items:
            self._visit_list(v, depth + 1, func)
            return

        if depth == len(self._paths) - 1:
            replacement = func(v)  # Includes None
            if self._modifiable and replacement:
                d[self._paths[depth]] = replacement
        else:
            if not v:
                return
            elif isinstance(v, dict):
                self._visit_dict(v, depth + 1, func)
            elif isinstance(v, list):
                self._visit_list(v, depth + 1, func)
            else:
                return  # Can't go further

    def _visit_list(self, lst: list, depth: int, func: PathVisitorCallback):
        if depth == len(self._paths):
            assert self._list_to_items
            for i, elem in enumerate(lst):
                replacement = func(elem)
                if self._modifiable and replacement:
                    lst[i] = replacement
        else:
            for i, elem in enumerate(lst):
                # Note: depth is not incremented in this case, since elements are at the same 'path depth' as the list
                if isinstance(elem, dict):
                    self._visit_dict(elem, depth, func)
                elif isinstance(elem, list):
                    self._visit_list(elem, depth, func)



"""
Build JobStats (returned to the client after job completion) - based mostly on the DataFrame of collected metrics from
the invoker and all workers.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import sys
from typing import Optional, Union, List, Dict
import pandas
import numpy as np
from pandas import DataFrame
from frocket.common.config import config
from frocket.common.dataset import DatasetPartsInfo, PartNamingMethod
from frocket.common.tasks.base import JobStats, JobDatasetStats, JobInvokerStats, TimingStats, JobWorkerStats
from frocket.invoker.metrics_frame import MetricsFrame, METRIC_NAME_COLUMN, METRIC_VALUE_COLUMN, METRIC_SOURCE_COLUMN
from frocket.common.metrics import MetricName, ComponentLabel, SUCCESS_LABEL, MetricLabelEnum, \
    WorkerStartupLabel, LoadFromLabel

logger = logging.getLogger(__name__)

TASK_COMPLETION_GRANULARITY_SECONDS = 0.25  # Data series of task success over time is measured in this resolution
TIMING_PERCENTILES = [float(pct) for pct in config.get('stats.timing.percentiles').split(',')]
MIN_METRICS_FOR_PERCENTILES = 20  # Below this sample count, don't return percentiles
MIN_METRICS_FOR_99_PERCENTILE = 100  # Below this count, don't return 99th percentile
# List of keys to pull from Pandas' describe()
TIMING_DESCRIBE_KEYS = ['min', 'mean', 'max'] + [f"{int(pct*100)}%" for pct in TIMING_PERCENTILES]


def build_stats(frame: MetricsFrame, parts_info: DatasetPartsInfo = None) -> JobStats:
    df = frame.dataframe
    if df is None:  # In job failure cases
        return JobStats()

    if parts_info:
        ds_stats = JobDatasetStats(total_size=parts_info.total_size, parts=parts_info.total_parts)
    else:
        ds_stats = None

    # Invoker stats
    all_task_rows_df = _filter_by_label(df, ComponentLabel.WORKER)
    successful_task_rows_df = _filter_by_success(all_task_rows_df)
    total_tasks = _count_tasks(all_task_rows_df)
    failed_tasks = total_tasks - _count_tasks(successful_task_rows_df)

    invoker_stats = JobInvokerStats(
        enqueue_time=_sum_value(df, MetricName.ASYNC_ENQUEUE_SECONDS, single_value=True),
        poll_time=_sum_value(df, MetricName.ASYNC_POLL_SECONDS, single_value=True),
        total_tasks=total_tasks,
        failed_tasks=failed_tasks,
        task_success_over_time=_task_success_over_time(successful_task_rows_df)
        # TODO backlog add: lost_task_retries as counted by the invoker; support sync. invokers?
    )

    # Worker stats
    worker_stats = JobWorkerStats(
        cold_tasks=_count_tasks(_filter_by_label(successful_task_rows_df, WorkerStartupLabel.COLD)),
        warm_tasks=_count_tasks(_filter_by_label(successful_task_rows_df, WorkerStartupLabel.WARM)),
        scanned_rows=_sum_value(successful_task_rows_df, MetricName.SCANNED_ROWS, as_int=True),
        scanned_groups=_sum_value(successful_task_rows_df, MetricName.SCANNED_GROUPS, as_int=True),
        cache=_cache_performance(successful_task_rows_df),
        invoke_latency=_timing_stats(successful_task_rows_df, MetricName.INVOKE_TO_RUN_SECONDS),
        load_time=_timing_stats(successful_task_rows_df, MetricName.TASK_TOTAL_LOAD_SECONDS),
        total_time=_timing_stats(successful_task_rows_df, MetricName.TASK_TOTAL_RUN_SECONDS)
        # TODO backlog add: loaded_column_types - mapping of column type to count, which affects load time
    )

    job_stats = JobStats(
        total_time=_sum_value(df, MetricName.INVOKER_TOTAL_SECONDS, single_value=True),
        cost=_total_cost(df),
        dataset=ds_stats,
        invoker=invoker_stats,
        worker=worker_stats)
    return job_stats


def _task_success_over_time(task_rows_df: DataFrame) -> Dict[float, int]:
    """Return a sparse series of data points - for each time slot (e.g. 0.25 secs) since the job started, return how
    many tasks completed successfully in that slot. Non-cumulative, does not include zeros."""
    task_duration_rows = _filter_by_metrics(
        task_rows_df, metrics=[MetricName.INVOKE_TO_RUN_SECONDS, MetricName.TASK_TOTAL_RUN_SECONDS])
    task_durations = task_duration_rows.groupby(METRIC_SOURCE_COLUMN)[METRIC_VALUE_COLUMN].sum()
    quantized_task_durations = \
        np.ceil(task_durations / TASK_COMPLETION_GRANULARITY_SECONDS) * TASK_COMPLETION_GRANULARITY_SECONDS
    return quantized_task_durations.value_counts().sort_index().to_dict()


def _cache_performance(task_rows_df: DataFrame) -> Dict[str, int]:
    return {
        # Note the 'source' is always the case for locally-loaded files, in which case caching is N/A.
        'source': _count_tasks(_filter_by_label(task_rows_df, LoadFromLabel.SOURCE)),
        'diskCache': _count_tasks(_filter_by_label(task_rows_df, LoadFromLabel.DISK_CACHE))
    }


def _sum_value(df: DataFrame, metric: MetricName,
               single_value: bool = False,
               as_int: bool = False) -> Union[float, int, None]:
    df = _filter_by_metrics(df, metric)
    if single_value:
        assert len(df) <= 1
    if df.empty:
        return None
    else:
        values_sum = df[METRIC_VALUE_COLUMN].sum()
        return int(values_sum) if as_int else float(values_sum)


def _count(df: DataFrame, metric: MetricName) -> int:
    return _filter_by_metrics(df, metric)[METRIC_VALUE_COLUMN].count()


def _timing_stats(task_rows_df: DataFrame, metric: MetricName) -> TimingStats:
    values_df = _filter_by_metrics(task_rows_df, metric)[METRIC_VALUE_COLUMN]
    if len(values_df) < MIN_METRICS_FOR_PERCENTILES:
        percentiles = [0.5]
    else:
        percentiles = TIMING_PERCENTILES
        if len(values_df) < MIN_METRICS_FOR_99_PERCENTILE:
            percentiles = [pct for pct in percentiles if pct < 0.99]

    raw_stats = values_df.describe(percentiles=percentiles).to_dict()
    return {k: v for k, v in raw_stats.items()
            if k in TIMING_DESCRIBE_KEYS and not np.isnan(v)}


def _filter_by_metrics(df: DataFrame, metrics: Union[MetricName, List[MetricName]]) -> DataFrame:
    if type(metrics) is MetricName:
        return df[df[METRIC_NAME_COLUMN] == metrics.name]
    else:
        return df[df[METRIC_NAME_COLUMN].isin([m.name for m in metrics])]


def _filter_by_label(df: DataFrame, label: MetricLabelEnum) -> DataFrame:
    return df[df[label.label_name] == label.label_value.lower()]


def _filter_by_success(df: DataFrame, value: bool = True) -> DataFrame:
    return df[df[SUCCESS_LABEL] == str(value)]


def _count_tasks(task_rows_df: DataFrame) -> int:
    """Each task attempt (e.g. task index 117, attempt 2) has a unique name in the source column, which ofc appears in
    multiple rows. This count the unique count of task attempt IDs in the given DF."""
    return task_rows_df[METRIC_SOURCE_COLUMN].nunique()


def _total_cost(df: DataFrame) -> Optional[float]:
    cost_reports_df = _filter_by_metrics(df, MetricName.COST_DOLLARS)
    num_reports = len(cost_reports_df)
    if num_reports == 0:
        logger.debug(f"Total cost: no metrics found")
        return None
    else:
        total_cost = float(cost_reports_df[METRIC_VALUE_COLUMN].sum())
        logger.debug(f"Total cost: ${total_cost:.6f} (sum of {num_reports} metric reports)")
        return total_cost


# Stand-alone testing
if __name__ == "__main__":
    config.init_logging(force_level=logging.DEBUG, force_console_output=True)
    filename = config.get('metrics.export.lastrun', None)
    if not filename:
        sys.exit('No lastrun file defined')

    df = pandas.read_parquet(filename)
    dummy_frame = MetricsFrame([])
    dummy_frame._df = df
    dummy_parts_info = DatasetPartsInfo(naming_method=PartNamingMethod.LIST, total_parts=4, total_size=1024)
    build_stats(dummy_frame, dummy_parts_info)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import os
import subprocess
from pathlib import Path
from typing import List, Union, cast
from frocket.cli import LOG_LINE_PREFIX
from frocket.common.config import config
from frocket.common.helpers.utils import timestamped_uuid, memoize
from frocket.invoker import invoker_api
from tests.utils.base_test_utils import temp_filename
from tests.utils.dataset_utils import new_test_dataset, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, \
    DEFAULT_GROUP_COUNT, TestColumn, datafile_schema
# noinspection PyUnresolvedReferences
from tests.utils.mock_s3_utils import mock_s3_env_variables
# noinspection PyUnresolvedReferences
from tests.utils.redis_fixture import init_test_redis_settings, get_test_redis_env_variables

CLI_MAIN = 'frocket.cli'


@memoize
def subprocess_env():
    full_env = {**os.environ, **get_test_redis_env_variables(), **mock_s3_env_variables()}
    print("Env variables for sub-process:", full_env)
    return full_env


def parse_cli_output(stdout_text: str) -> List[str]:
    assert type(stdout_text) is str
    log_min_failure_level = logging.WARNING
    raw_lines = stdout_text.split('\n')
    lines = []

    for raw_line in raw_lines:
        raw_line = raw_line.strip()
        if not raw_line:  # Empty line
            continue
        if raw_line.startswith(LOG_LINE_PREFIX):  # Detect log line, and assert by severity. Log lines are not returned
            log_level = raw_line.split(' ')[1]
            assert logging.getLevelName(log_level) < log_min_failure_level
        else:
            lines.append(raw_line)
    return lines


def call_cli(cmd: str,
             args: List[str] = None,
             expected_success: bool = True,
             as_single_json: bool = False,
             api_result_format: bool = False) -> Union[List[str], dict]:
    if args is None:
        args = []
    full_args = ['python', '-m', CLI_MAIN, '--notrim', '--nocolor', '--nopretty']
    if not expected_success:
        full_args += ['--loglevel', 'critical']
    full_args += [cmd, *args]

    result = subprocess.run(full_args, capture_output=True, text=True, env=subprocess_env())
    print(f'Expected success: {expected_success}, actual return code: {result.returncode}, args: {result.returncode}')
    print('stdout:', result.stdout)
    print('stderr:', result.stderr)

    assert (result.returncode == 0) == expected_success
    lines = parse_cli_output(result.stdout)
    if as_single_json:
        assert len(lines) == 1
        d = json.loads(lines[0])
        if api_result_format:
            assert d['success'] == expected_success
        return d
    else:
        return lines


def test_invalid_cmd():
    call_cli('whatthehell', expected_success=False)
    call_cli('whattheheck', args=['blah'] * 100, expected_success=False)


def test_list():
    cli_datasets = cast(dict, call_cli('list', as_single_json=True))
    expected_ds_names = sorted([ds.id.name for ds in invoker_api.list_datasets()])
    assert sorted([ds['name'] for ds in cli_datasets]) == expected_ds_names


def test_config():
    cli_config = cast(dict, call_cli('config', as_single_json=True))
    cli_config.pop('log.format')

    expected_config = dict(config)
    expected_config.pop('log.format')
    assert cli_config == expected_config


def test_register_unregister():
    with new_test_dataset(parts=2) as test_ds:
        s3path = test_ds.copy_to_s3()
        name = timestamped_uuid(prefix='testcli-')
        call_cli('register', args=[name, s3path, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN],
                 as_single_json=True, api_result_format=True)
        call_cli('unregister', args=[name], as_single_json=True, api_result_format=True)


def test_register_fail():
    with new_test_dataset(parts=2) as test_ds:
        s3path = test_ds.copy_to_s3()
        name = timestamped_uuid(prefix='testcli-')
        # Invalid path
        call_cli('register', args=[name, s3path + '/blah', DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN],
                 expected_success=False, as_single_json=True, api_result_format=True)
        # Invalid column name
        call_cli('register', args=[name, s3path, DEFAULT_GROUP_COLUMN + '-not', DEFAULT_TIMESTAMP_COLUMN],
                 expected_success=False, as_single_json=True, api_result_format=True)
        # Invalid validation mode
        call_cli('register',
                 args=[name, s3path, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, '--validation', 'nosuchmode'],
                 expected_success=False,
                 as_single_json=False,
                 api_result_format=False)
        # Should work (with optional flags)
        call_cli('register', args=[name, s3path, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN,
                                   '--validation', 'FIRST_LAST', '--skip-uniques'],
                 as_single_json=True,
                 api_result_format=True)
        call_cli('unregister', args=[name], as_single_json=True, api_result_format=True)


def test_query_modes():
    with new_test_dataset(1) as test_ds:
        s3path = test_ds.copy_to_s3()
        name = timestamped_uuid(prefix='testcli-')
        call_cli('register', args=[name, s3path, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN],
                 as_single_json=True, api_result_format=True)

        empty_query = '{}'
        # Invalid cases
        call_cli('run', args=[], expected_success=False)
        call_cli('run', args=[name, '--empty', '--string', 'empty_query'], expected_success=False)
        call_cli('run', args=[name, '--file', 'blahblah'], expected_success=False)
        call_cli('run', args=[name, '--string', '"{{{}"'], expected_success=False)
        call_cli('run', args=["notreally...", '--empty'], expected_success=False)

        # Valid empty query in all modes
        empty_result1 = call_cli('run', args=[name, '--empty'], as_single_json=True, api_result_format=True)
        empty_result2 = call_cli('run', args=[name, '--string', empty_query],
                                 as_single_json=True, api_result_format=True)
        empty_query_file = temp_filename()
        Path(empty_query_file).write_text('{}', encoding='utf-8')
        empty_result3 = call_cli('run', args=[name, '--file', empty_query_file],
                                 as_single_json=True, api_result_format=True)
        os.remove(empty_query_file)

        for res in [empty_result1, empty_result2, empty_result3]:
            assert res['query']['matchingGroups'] == DEFAULT_GROUP_COUNT
            assert res['stats']['invoker']['totalTasks'] == 1
            assert res['query']['aggregations'] is None

        # Valid non-empty queries
        q1 = {'query': {'aggregations': [{'column': TestColumn.bool.value, 'type': 'count', 'name': 'aggr_string'}]}}
        result = call_cli('run', args=[name, '--string', json.dumps(q1)], as_single_json=True, api_result_format=True)
        assert result['query']['aggregations'][0]['name'] == 'aggr_string'

        q2 = {'query': {'aggregations': [{'column': TestColumn.bool.value, 'type': 'count', 'name': 'aggr_file'}]}}
        q2file = temp_filename()
        Path(q2file).write_text(json.dumps(q2), encoding='utf-8')
        result = call_cli('run', args=[name, '--file', q2file], as_single_json=True, api_result_format=True)
        assert result['query']['aggregations'][0]['name'] == 'aggr_file'
        os.remove(q2file)

        # Cleanup
        call_cli('unregister', args=[name, '--force'], as_single_json=True, api_result_format=True)


def test_info():
    with new_test_dataset(1) as test_ds:
        s3path = test_ds.copy_to_s3()
        name = timestamped_uuid(prefix='testcli-')
        call_cli('register', args=[name, s3path, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN],
                 as_single_json=True, api_result_format=True)

        def get_info(full: bool = False):
            args = [name]
            if full:
                args += ['--full']
            result = call_cli('info', args=args)
            assert len(result) == 3
            basic_info = json.loads(result[0])
            parts_info = json.loads(result[1])
            schema = json.loads(result[2])
            return basic_info, parts_info, schema

        basic_info, parts_info, short_schema = get_info(full=False)
        assert basic_info['id']['name'] == name
        assert basic_info['basepath'] == s3path
        assert parts_info['filenames'] == test_ds.expected_parts.filenames
        assert short_schema == json.loads(datafile_schema().to_json())

        basic_info, parts_info, full_schema = get_info(full=True)
        assert basic_info['id']['name'] == name
        assert parts_info['filenames'] == test_ds.expected_parts.filenames
        # print(full_schema)
        group_col_data = full_schema['columns'][DEFAULT_GROUP_COLUMN]
        assert group_col_data['name'] == DEFAULT_GROUP_COLUMN
        assert group_col_data['colattrs']['numericMin'] == 0.0
        assert group_col_data['colattrs']['numericMax'] == DEFAULT_GROUP_COUNT - 1

        call_cli('unregister', args=[name], as_single_json=True, api_result_format=True)



import sys
import time
import argparse
from pathlib import Path
from contextlib import contextmanager
import pandas as pd
from pandas import DataFrame

EVENTS_FILE = 'events.csv'
PROPS_FILE_1 = 'item_properties_part1.csv'
PROPS_FILE_2 = 'item_properties_part2.csv'
INPUT_FILENAMES = {EVENTS_FILE, PROPS_FILE_1, PROPS_FILE_2}
ITEM_PROPERTY_COLUMNS = {'categoryid', 'available', '790', '888'}
EXPECTED_EVENT_COUNT = 2_500_516


def progress_msg(msg: str):
    print(f"\033[33m{msg}\033[0m")  # Yellow, just yellow


@contextmanager
def timed(caption: str):
    start = time.time()
    yield
    total = time.time() - start
    print(f"Time to {caption}: {total:.3f} seconds")


# Read item properties files, filter for relevant columns and 'pivot' its structure from rows to columns
def read_item_props(filepath: Path) -> DataFrame:
    df = pd.read_csv(filepath)
    df = df[df['property'].isin(ITEM_PROPERTY_COLUMNS)]
    first_value_per_item = df.groupby(["itemid", "property"])["value"].first()
    df = first_value_per_item.to_frame()
    df = df.unstack(level=-1)
    df.columns = df.columns.droplevel(0)
    return df


def ingest(path: Path):
    with timed("read & transform item properties of all products"):
        item_props_tempfile = path / "item_props.parquet"
        if item_props_tempfile.exists():
            progress_msg(f"Reading item properties from cached file {item_props_tempfile}")
            item_props_df = pd.read_parquet(item_props_tempfile)
        else:
            progress_msg("Reading item properties... (this takes a bit)")
            item_props_df1 = read_item_props(path / PROPS_FILE_1)
            item_props_df2 = read_item_props(path / PROPS_FILE_2)
            item_props_df = item_props_df1.combine_first(item_props_df2)
            progress_msg(f"Storing item properties to {item_props_tempfile} for faster re-runs...")
            item_props_df.to_parquet(item_props_tempfile)

    with timed("read & transform user events"):
        progress_msg("Reading user events...")
        events = pd.read_csv(path / EVENTS_FILE)
        progress_msg("Joining events with item properties...")
        events = pd.merge(events, item_props_df, how='inner', on='itemid')

        progress_msg("Making columns more queryable...")
        events['price'] = events['790'].str[1:].astype(float) / 1000
        events.drop(columns=['790'], inplace=True)
        events['available'] = events['available'].astype(int).astype(bool)
        events['categoryid'] = events['categoryid'].astype('category')
        events['event'] = events['event'].astype('category')
        events.rename(columns={'888': 'cryptic_attrs'}, inplace=True)
        progress_msg("Storing 'cryptic_attrs' also as categorical column 'cryptic_attrs_cat'...")
        events['cryptic_attrs_cat'] = events['cryptic_attrs'].astype('category')
        events.reset_index(drop=True)

    progress_msg("Excerpt from final DataFrame:")
    print(events)
    progress_msg("Columns types (a.k.a. dtypes):")
    print(events.dtypes)
    progress_msg("Breakdown of event types:")
    print(events['event'].value_counts())

    if len(events) != EXPECTED_EVENT_COUNT:
        progress_msg(f"WARNING: Expected {EXPECTED_EVENT_COUNT} events, but final DataFrame has {len(events)}")

    output_file = path / 'retailrocket.parquet'
    events.to_parquet(output_file)
    col_memory_sizes = (events.memory_usage(deep=True) / 1024 ** 2).round(decimals=2)
    progress_msg(f'Size of DataFrame columns in memory (in MB):')
    print(col_memory_sizes)
    progress_msg(f"==> Saved output file to: {output_file}, size: {output_file.stat().st_size / 1024 ** 2:.1f}MB")

    with timed("load file - all columns"):
        pd.read_parquet(output_file)

    with timed("load file - just the 'cryptic_attrs' column"):
        pd.read_parquet(output_file, columns=['cryptic_attrs'])

    with timed("load file - just the 'cryptic_attrs_cat' column"):
        pd.read_parquet(output_file, columns=['cryptic_attrs_cat'])

    with timed("load file - all columns *except* these two"):
        cols = [col for col in events.dtypes.index
                if col not in ['cryptic_attrs', 'cryptic_attrs_cat']]
        pd.read_parquet(output_file, columns=cols)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Ingest RetailRocket dataset (to download: https://www.kaggle.com/retailrocket/ecommerce-dataset/)')
    parser.add_argument(
        'path', type=str,
        help='Directory where downloaded dataset files are found and output file will be written')
    args = parser.parse_args()

    path = Path(args.path)
    if not path.exists() or not path.is_dir():
        sys.exit(f'No such directory: {path}')
    files_in_path = {f.name for f in path.iterdir()}
    if not files_in_path >= INPUT_FILENAMES:
        sys.exit(f'Missing one or more input files: {INPUT_FILENAMES}')
    ingest(path)



"""
lambda_handler() in this module is the AWS Lambda's defined entrypoint.
There's minimal code here that's Lambda-specific (== a good thing).
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import cast
from frocket.common.serializable import Envelope
from frocket.common.tasks.base import BaseTaskRequest
from frocket.common.metrics import MetricsBag, WorkerStartupLabel, ComponentLabel
from frocket.worker.impl.aws_lambda_metrics import AwsLambdaMetricsProvider
from frocket.worker.runners.base_task_runner import BaseTaskRunner, TaskRunnerContext
from frocket.common.config import config
from frocket.worker.runners.registered_runners import REGISTERED_RUNNERS

config.init_lambda_logging()  # Adapted to the logger being already-inited by the Lambda runtime
logger = logging.getLogger(__name__)

# This flag only set when a new Lambda instance is cold-started. Warm lambdas would go straight to the handler function.
cold_start_flag = True


def is_cold_start():
    global cold_start_flag
    if cold_start_flag:
        cold_start_flag = False  # For next invocation
        return True
    else:
        return False


def init_task_metrics(lambda_context) -> MetricsBag:
    metrics = MetricsBag(component=ComponentLabel.WORKER,
                         env_metrics_provider=AwsLambdaMetricsProvider(lambda_context))
    if is_cold_start():
        metrics.set_label_enum(WorkerStartupLabel.COLD)
    else:
        metrics.set_label_enum(WorkerStartupLabel.WARM)
    return metrics


def lambda_handler(event, context):
    metrics = init_task_metrics(context)
    # The event JSON was already parsed to dict by the Lambda runtime -
    # now read from that dict that actual task request object
    envelope = Envelope.from_dict(event)
    req = cast(BaseTaskRequest, envelope.open(expected_superclass=BaseTaskRequest))
    logger.info(f"Got request: {req}")

    result = None
    should_run, reject_reason = BaseTaskRunner.should_run(req)
    if should_run:
        runner_class = REGISTERED_RUNNERS[type(req)]
        runner = runner_class(req, TaskRunnerContext(metrics))
        result = runner.run()

    """
    A note about the Lambda response: unlike most request/response Lambdas, Funnel Rocket's invoker does not rely on the
    function's result coming from the Lambda directly (as it's invoked async.) but rather always through the datastore.
    The retry mechanism is also based on polling the tasks' status and result payload in the datastore, hence the 
    Lambda itself should not normally return a non-200 status (unless it crashed unexpectedly), and the Lambda should
    be configured to have no retries at the AWS level.
    """

    lambda_response = {
        'statusCode': 200,
    }

    # Getting the result object in the Lambda response is still useful for manual testing
    if logger.isEnabledFor(logging.DEBUG):
        if result:
            lambda_response['result'] = result.to_json()
        else:
            lambda_response['reject_reason'] = reject_reason
    return lambda_response



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from enum import Enum, auto
from frocket.common.config import config
from frocket.invoker.base_invoker import BaseInvoker
from frocket.invoker.jobs.job import Job
from frocket.invoker.impl.aws_lambda_invoker import AwsLambdaInvoker
from frocket.invoker.impl.work_queue_invoker import WorkQueueInvoker

logger = logging.getLogger(__name__)


class InvocationType(Enum):
    WORK_QUEUE = auto()
    AWS_LAMBDA = auto()


INVOKER_CLASSES = {
    InvocationType.WORK_QUEUE: WorkQueueInvoker,
    InvocationType.AWS_LAMBDA: AwsLambdaInvoker
}


def new_invoker(request_builder: Job) -> BaseInvoker:
    invoker_type = InvocationType[config.get("invoker").upper()]
    invoker_class = INVOKER_CLASSES[invoker_type]
    logger.info(f"Creating invoker type: {invoker_class.__name__}, for request builder type: {type(request_builder)}")
    return invoker_class(request_builder)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import os
from typing import Dict, List
import pandas
import pytest
from frocket.common.dataset import DatasetPartId, DatasetId
from frocket.common.helpers.utils import timestamped_uuid
from frocket.common.metrics import MetricsBag, ComponentLabel, LoadFromLabel
from tests.utils.dataset_utils import new_test_dataset, are_test_dfs_equal, clean_loader_cache, TestColumn, BASE_TIME, \
    TIME_SHIFT
from frocket.worker.runners.part_loader import shared_part_loader, FilterPredicate
from tests.utils.mock_s3_utils import SKIP_S3_TESTS


def test_local_non_cached():
    with clean_loader_cache():
        with new_test_dataset(4) as test_ds:
            dataset_id = DatasetId.now(timestamped_uuid())
            original_df = pandas.read_parquet(test_ds.fullpath_files[0])
            file_id = DatasetPartId(dataset_id, path=test_ds.fullpath_files[0], part_idx=0)

            metrics = MetricsBag(component=ComponentLabel.WORKER)
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics)
            assert are_test_dfs_equal(original_df, loaded_df)
            assert metrics.label_value(LoadFromLabel) == LoadFromLabel.SOURCE.label_value

            metrics = MetricsBag(component=ComponentLabel.WORKER)
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics)
            assert are_test_dfs_equal(original_df, loaded_df)
            assert metrics.label_value(LoadFromLabel) == LoadFromLabel.SOURCE.label_value


def test_needed_columns():
    with clean_loader_cache():
        with new_test_dataset(1) as test_ds:
            dataset_id = DatasetId.now(timestamped_uuid())
            original_df = pandas.read_parquet(test_ds.fullpath_files[0])
            file_id = DatasetPartId(dataset_id, path=test_ds.fullpath_files[0], part_idx=0)
            cols = [TestColumn.int_64_userid, TestColumn.str_userid, TestColumn.str_all_none,
                    TestColumn.float_32, TestColumn.str_category_many]

            metrics = MetricsBag(component=ComponentLabel.WORKER)
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, needed_columns=cols)
            assert are_test_dfs_equal(original_df[cols], loaded_df)
            assert metrics.label_value(LoadFromLabel) == LoadFromLabel.SOURCE.label_value

            metrics = MetricsBag(component=ComponentLabel.WORKER)
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, needed_columns=cols)
            assert are_test_dfs_equal(original_df[cols], loaded_df)
            assert metrics.label_value(LoadFromLabel) == LoadFromLabel.SOURCE.label_value

            metrics = MetricsBag(component=ComponentLabel.WORKER)
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, needed_columns=cols,
                                                            load_as_categoricals=['str_userid'])
            assert loaded_df['str_userid'].dtype == 'category'
            assert are_test_dfs_equal(original_df[cols], loaded_df)


def test_filters():
    # Note that we're not testing Arrow's filter functionality, but rather that we're using it as intended
    with clean_loader_cache():
        with new_test_dataset(1) as test_ds:
            dataset_id = DatasetId.now(timestamped_uuid())
            original_df = pandas.read_parquet(test_ds.fullpath_files[0])
            file_id = DatasetPartId(dataset_id, path=test_ds.fullpath_files[0], part_idx=0)

            # All rows should return
            metrics = MetricsBag(component=ComponentLabel.WORKER)
            filters = [FilterPredicate(TestColumn.int_64_ts.value, '>=', BASE_TIME),
                       FilterPredicate(TestColumn.int_64_ts.value, '<=', BASE_TIME + TIME_SHIFT)]
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, filters=filters)
            assert are_test_dfs_equal(original_df, loaded_df)

            # Only rows with exact match should return
            metrics = MetricsBag(component=ComponentLabel.WORKER)
            filters = [FilterPredicate(TestColumn.int_64_ts.value, '==', BASE_TIME)]
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, filters=filters)
            assert are_test_dfs_equal(original_df.loc[original_df[TestColumn.int_64_ts] == BASE_TIME], loaded_df)

            # Now test a string column filter + specific needed columns
            metrics = MetricsBag(component=ComponentLabel.WORKER)
            filters = [FilterPredicate(TestColumn.str_userid.value, '==', '0')]
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, filters=filters,
                                                            needed_columns=[TestColumn.bool, TestColumn.str_userid])
            # In the loaded DF, matching rows have 0..n index (regardless of their "original" position)
            expected_df = original_df[[TestColumn.bool, TestColumn.str_userid]]. \
                loc[original_df[TestColumn.str_userid] == '0']. \
                reset_index(drop=True)
            assert are_test_dfs_equal(expected_df, loaded_df)

            # No rows match
            metrics = MetricsBag(component=ComponentLabel.WORKER)
            filters = [FilterPredicate(TestColumn.int_64_ts.value, '<', BASE_TIME)]
            loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics, filters=filters)
            assert len(loaded_df) == 0


def validate_load(file_id: DatasetPartId, original_df: pandas.DataFrame, expected_load_from: LoadFromLabel):
    metrics = MetricsBag(component=ComponentLabel.WORKER)
    loaded_df = shared_part_loader().load_dataframe(file_id, metrics=metrics)
    assert are_test_dfs_equal(original_df, loaded_df)
    assert metrics.label_value(LoadFromLabel) == expected_load_from.label_value


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_remote_caching():
    with clean_loader_cache():
        with new_test_dataset(4) as test_ds:
            s3path = test_ds.copy_to_s3(timestamped_uuid())

            ds_name = timestamped_uuid()
            dataset_id = DatasetId.now(ds_name)
            original_df1 = pandas.read_parquet(test_ds.fullpath_files[0])
            original_df2 = pandas.read_parquet(test_ds.fullpath_files[1])
            partid1 = DatasetPartId(dataset_id, path=f"{s3path}/{test_ds.basename_files[0]}", part_idx=0)
            partid2 = DatasetPartId(dataset_id, path=f"{s3path}/{test_ds.basename_files[1]}", part_idx=1)

            validate_load(partid1, original_df1, LoadFromLabel.SOURCE)
            for i in range(3):
                validate_load(partid1, original_df1, LoadFromLabel.DISK_CACHE)

            validate_load(partid2, original_df2, LoadFromLabel.SOURCE)
            for i in range(3):
                validate_load(partid2, original_df2, LoadFromLabel.DISK_CACHE)
                validate_load(partid1, original_df1, LoadFromLabel.DISK_CACHE)

            new_dataset_id = DatasetId.now(ds_name)
            new_fileid1 = DatasetPartId(new_dataset_id, path=partid1.path, part_idx=partid1.part_idx)
            validate_load(new_fileid1, original_df1, LoadFromLabel.SOURCE)
            validate_load(new_fileid1, original_df1, LoadFromLabel.DISK_CACHE)
            # NOTE: For now, cache doesn't auto-clear when newer dataset revision is found
            validate_load(partid1, original_df1, LoadFromLabel.DISK_CACHE)

            assert shared_part_loader().cache_len == 3


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_cache_size():
    num_parts = 4
    with new_test_dataset(num_parts) as test_ds:
        name = timestamped_uuid()
        s3path = test_ds.copy_to_s3(name)
        dataset_id = DatasetId.now(name)
        original_dfs = [pandas.read_parquet(test_ds.fullpath_files[i])
                        for i in range(num_parts)]
        part_ids = [DatasetPartId(dataset_id, path=f"{s3path}/{test_ds.basename_files[i]}", part_idx=i)
                    for i in range(num_parts)]

        with clean_loader_cache(size_mb=0):
            for i in range(num_parts):
                validate_load(part_ids[i], original_dfs[i], LoadFromLabel.SOURCE)
                validate_load(part_ids[i], original_dfs[i], LoadFromLabel.SOURCE)

            for i in range(num_parts):
                validate_load(part_ids[i], original_dfs[i], LoadFromLabel.SOURCE)

            # With no cache, the most recent file loaded was still downloaded locally
            # TODO reconsider this behavior
            assert shared_part_loader().cache_len == 1
            filesize_mb = os.path.getsize(test_ds.fullpath_files[-1]) / (1024 ** 2)
            assert shared_part_loader().cache_current_size_mb == filesize_mb

        # Constrain cache size to contain about half the dataset files, +/-1
        # This assumes all test files are about the same size.
        assert shared_part_loader().cache_len == 0
        half_ds_size_bytes = sum([os.path.getsize(test_ds.fullpath_files[i]) for i in range(num_parts - 2)])
        half_ds_size_mb = half_ds_size_bytes / (1024 ** 2)

        with clean_loader_cache(size_mb=half_ds_size_mb):
            for i in range(num_parts):
                validate_load(part_ids[i], original_dfs[i], LoadFromLabel.SOURCE)
            # Most recently used should be cached
            validate_load(part_ids[-1], original_dfs[-1], LoadFromLabel.DISK_CACHE)

            # Going again from the least-recently used, they should all be evicted by now
            # while parts re-loaded replace the cache content
            for i in range(num_parts):
                validate_load(part_ids[i], original_dfs[i], LoadFromLabel.SOURCE)

            # Cache should hold up to about half the files
            assert 0 < shared_part_loader().cache_len <= (num_parts / 2 + 1)

            # All parts now in cache should be "valid candidates" from the same dataset.
            # Basically, using the get_cached_candidates() to extract cache contents
            parts_in_cache = shared_part_loader().get_cached_candidates(dataset_id)
            assert len(parts_in_cache) == shared_part_loader().cache_len
            # Validate that it's in fact the most-recently used parts that are in that list
            parts_in_cache = sorted(parts_in_cache, key=lambda part: part.part_idx)
            assert parts_in_cache == part_ids[-len(parts_in_cache):]


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_get_candidates():
    # Creating 3 different DatasetIds (two of them with same name but different reg. data),
    # and ensure that parts are cached as different files (although source path in S3 is similar),
    # and that candidates are returned for the relevant ds ID only
    num_parts = 2
    with new_test_dataset(num_parts) as test_ds:
        s3path = test_ds.copy_to_s3()
        original_dfs = [pandas.read_parquet(test_ds.fullpath_files[i])
                        for i in range(num_parts)]

        dsid_1a = DatasetId.now("ds1")
        dsid_1b = DatasetId(name=dsid_1a.name,
                            registered_at=dsid_1a.registered_at + datetime.timedelta(seconds=1))
        dsid_2 = DatasetId.now("ds2")
        dsid_to_parts: Dict[DatasetId, List[DatasetPartId]] = {}

        def add_parts(dsid: DatasetId):
            dsid_to_parts[dsid] = [DatasetPartId(dsid, path=f"{s3path}{test_ds.basename_files[i]}", part_idx=i)
                                   for i in range(num_parts)]
        add_parts(dsid_1a)
        add_parts(dsid_1b)
        add_parts(dsid_2)

        with clean_loader_cache():
            for parts in dsid_to_parts.values():
                for part in parts:
                    validate_load(part, original_dfs[part.part_idx], LoadFromLabel.SOURCE)

            for dsid, parts in dsid_to_parts.items():
                dsid_copy = DatasetId(name=dsid.name, registered_at=dsid.registered_at)
                candidates = shared_part_loader().get_cached_candidates(dsid_copy)
                candidates = sorted(candidates, key=lambda k: k.part_idx)
                assert candidates == parts



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import os
import shutil
from typing import List, cast
import pytest
import pandas as pd
from tests.utils.mock_s3_utils import SKIP_S3_TESTS
from tests.utils.task_and_job_utils import registration_job_invoker, build_registration_job, registration_job
from tests.utils.dataset_utils import new_test_dataset, STR_CAT_MANY_WEIGHTS, TestColumn, DEFAULT_GROUP_COUNT
from tests.utils.base_test_utils import temp_filename, SKIP_SLOW_TESTS
from frocket.common.helpers.utils import bytes_to_ndarray
from frocket.common.tasks.registration import DatasetValidationMode,  \
    RegistrationTaskResult, RegistrationJobResult
from frocket.datastore.registered_datastores import get_datastore, get_blobstore
from frocket.invoker.jobs.registration_job import VALIDATION_MAX_SAMPLES, \
    VALIDATION_SAMPLE_RATIO, CATEGORICAL_TOP_COUNT
# noinspection PyUnresolvedReferences
from tests.utils.redis_fixture import init_test_redis_settings


# TODO create dataset and tear it down nicer
def test_local_single_file_discovery():
    with new_test_dataset(1) as test_ds:
        with registration_job(test_ds, DatasetValidationMode.SINGLE) as job:
            job.prerun()

            assert job.parts == test_ds.expected_parts
            assert job.sampled_parts == test_ds.expected_part_ids(job.dataset.id)

        with registration_job(test_ds, DatasetValidationMode.SAMPLE, uniques=True) as job:
            job.prerun()
            assert job.parts == test_ds.expected_parts
            assert job.sampled_parts == test_ds.expected_part_ids(job.dataset.id)

        with registration_job(test_ds, DatasetValidationMode.FIRST_LAST, uniques=True, pattern="*t") as job:
            job.prerun()
            assert job.parts == test_ds.expected_parts
            assert job.sampled_parts == test_ds.expected_part_ids(job.dataset.id)

        with open(f"{test_ds.basepath}/touched", "w") as f:
            f.write("hello")

        with registration_job(test_ds, DatasetValidationMode.FIRST_LAST) as job:
            job.prerun()
            assert job.parts == test_ds.expected_parts
            assert job.sampled_parts == test_ds.expected_part_ids(job.dataset.id)


def do_test_multiple_local_files(parts: int):
    with new_test_dataset(parts) as test_ds:
        with registration_job(test_ds, DatasetValidationMode.SINGLE) as job:
            job.prerun()
            assert sorted(job.parts.filenames) == sorted(test_ds.expected_parts.filenames)
            assert job.parts.total_size == test_ds.expected_parts.total_size
            assert len(job.sampled_parts) == 1
            assert job.sampled_parts[0].path in test_ds.fullpath_files

        with registration_job(test_ds, DatasetValidationMode.FIRST_LAST) as job:
            job.prerun()
            assert sorted(job.parts.filenames) == sorted(test_ds.expected_parts.filenames)
            assert job.parts.total_size == test_ds.expected_parts.total_size
            assert len(job.sampled_parts) == 2
            sampled_fnames = sorted([sp.path for sp in job.sampled_parts])
            assert sampled_fnames[0] == sorted(test_ds.fullpath_files)[0]
            assert sampled_fnames[-1] == sorted(test_ds.fullpath_files)[-1]

        with registration_job(test_ds, DatasetValidationMode.SAMPLE) as job:
            job.prerun()
            assert sorted(job.parts.filenames) == sorted(test_ds.expected_parts.filenames)
            assert job.parts.total_size == test_ds.expected_parts.total_size
            assert 2 <= len(job.sampled_parts) <= len(test_ds.fullpath_files)
            assert len(job.sampled_parts) <= VALIDATION_MAX_SAMPLES
            sampled_fnames = sorted([sp.path for sp in job.sampled_parts])
            assert sampled_fnames[0] == sorted(test_ds.fullpath_files)[0]
            assert sampled_fnames[-1] == sorted(test_ds.fullpath_files)[-1]

            if len(test_ds.fullpath_files) > 2:
                expected_sample_count = math.floor(len(test_ds.fullpath_files) * VALIDATION_SAMPLE_RATIO)
                if expected_sample_count <= VALIDATION_MAX_SAMPLES:
                    assert len(sampled_fnames) == expected_sample_count
                else:
                    assert len(sampled_fnames) == VALIDATION_MAX_SAMPLES
                sampled_fnames = [sp.path for sp in job.sampled_parts]
                assert all([fname in test_ds.fullpath_files for fname in sampled_fnames])


def test_local_two_file_discovery():
    do_test_multiple_local_files(2)


# TODO what's the right marker
@pytest.mark.skipif(SKIP_SLOW_TESTS, reason="Skipping slow tests")
def test_local_many_file_discovery():
    do_test_multiple_local_files(20)
    do_test_multiple_local_files(120)


@pytest.mark.xfail(strict=True, raises=FileNotFoundError)
def test_local_missing_path():
    job = build_registration_job(basepath="blahblahblah", mode=DatasetValidationMode.SINGLE)
    job.prerun()


def test_nofiles_no_samples():
    basepath = temp_filename()
    os.mkdir(basepath)
    job = build_registration_job(basepath, mode=DatasetValidationMode.SINGLE)
    job.prerun()
    assert job.sampled_parts is None


@pytest.mark.xfail(strict=True)
def test_nofiles_failure():
    basepath = temp_filename()
    os.mkdir(basepath)
    job = build_registration_job(basepath, DatasetValidationMode.SINGLE)
    with registration_job_invoker(job) as invoker:
        invoker.prerun()
        invoker.enqueue_tasks()


def test_local_nondefault_pattern():
    num_parts = 4
    with new_test_dataset(num_parts, "blah-", ".123") as test_ds:
        with open(f"{test_ds.basepath}/single", "w") as f:
            f.write("herehere")

        mode = DatasetValidationMode.SINGLE
        with registration_job(test_ds, mode=mode) as job:
            job.prerun()
            assert job.sampled_parts is None

        with registration_job(test_ds, mode=mode, pattern="blahx*") as job:
            job.prerun()
            assert job.sampled_parts is None

        with registration_job(test_ds, mode=mode, pattern="bl*.123") as job:
            job.prerun()
            assert job.parts.total_parts == num_parts

        with registration_job(test_ds, mode=mode, pattern="single") as job:
            job.prerun()
            assert job.parts.total_parts == 1


def test_merged_schema():
    num_parts = 20
    with new_test_dataset(num_parts) as test_ds:
        with registration_job(test_ds, mode=DatasetValidationMode.SAMPLE, uniques=True) as job:
            with registration_job_invoker(job) as invoker:
                invoker.prerun()
                invoker.enqueue_tasks()
                assert invoker.num_tasks >= 2 and invoker.num_tasks == len(job.sampled_parts)

                task_results = cast(List[RegistrationTaskResult], invoker.run_tasks())
                paths = [tr.part_id.path for tr in task_results]
                assert sorted(paths) == sorted([sp.path for sp in job.sampled_parts])

                all_group_ids = []
                for res in task_results:
                    blob_content = get_blobstore().read_blob(res.group_ids_blob_id)
                    assert blob_content
                    uniques = bytes_to_ndarray(blob_content)
                    all_group_ids += list(uniques)
                assert len(all_group_ids) == DEFAULT_GROUP_COUNT * len(job.sampled_parts)
                assert len(set(all_group_ids)) == len(all_group_ids)

                final_job_status = invoker.complete()
                job_result = cast(RegistrationJobResult, invoker.build_result())

                result_ds = job_result.dataset
                assert get_datastore().dataset_info(result_ds.id.name) == result_ds

                result_schema = get_datastore().schema(result_ds)
                result_short_schema = get_datastore().short_schema(result_ds)

    assert task_results[0].dataset_schema.group_id_column == result_schema.group_id_column
    assert task_results[0].dataset_schema.timestamp_column == result_schema.timestamp_column
    assert task_results[0].dataset_schema.short().columns == result_short_schema.columns
    assert task_results[0].dataset_schema.short().source_categoricals == result_short_schema.source_categoricals
    assert task_results[0].dataset_schema.short().potential_categoricals == \
           result_short_schema.potential_categoricals

    min_ts = min([res.dataset_schema.columns[result_schema.timestamp_column].colattrs.numeric_min
                  for res in task_results])
    max_ts = max([res.dataset_schema.columns[result_schema.timestamp_column].colattrs.numeric_max
                  for res in task_results])

    assert result_short_schema.min_timestamp == min_ts
    assert result_short_schema.max_timestamp == max_ts

    attrs = result_schema.columns[TestColumn.str_category_many].colattrs
    assert attrs.categorical and len(attrs.cat_top_values) == CATEGORICAL_TOP_COUNT
    assert attrs.cat_top_values['0'] == STR_CAT_MANY_WEIGHTS[0]
    assert attrs.cat_top_values['1'] == STR_CAT_MANY_WEIGHTS[1]


def test_non_unique_ids():
    with new_test_dataset(2) as test_ds:
        # Copy first file over the second one - so they now have the same group IDs
        shutil.copyfile(test_ds.fullpath_files[0], test_ds.fullpath_files[-1])
        with registration_job(test_ds, mode=DatasetValidationMode.FIRST_LAST, uniques=True) as job:
            with registration_job_invoker(job) as invoker:
                invoker.prerun()
                assert len(job.sampled_parts) == 2
                invoker.enqueue_tasks()
                invoker.run_tasks()  # This checks that all *task* results have finished ok
                # The job fails (at the complete() phase) but does not crash
                final_job_status = invoker.complete(assert_success=False)
                assert not final_job_status.success
                result = invoker.build_result(assert_success=False)
                assert not result.success


def test_schemas_differ():
    with new_test_dataset(2) as test_ds:
        fname = test_ds.fullpath_files[-1]
        df = pd.read_parquet(fname)
        df['added_column'] = df[TestColumn.int_64_userid]
        df.to_parquet(fname)

        with registration_job(test_ds, mode=DatasetValidationMode.FIRST_LAST, uniques=True) as job:
            with registration_job_invoker(job) as invoker:
                invoker.prerun()
                assert len(job.sampled_parts) == 2
                invoker.enqueue_tasks()
                invoker.run_tasks()  # This checks that all *task* results have finished ok
                # The job fails (at the complete() phase) but does not crash
                final_job_status = invoker.complete(assert_success=False)
                assert not final_job_status.success
                result = invoker.build_result(assert_success=False)
                assert not result.success


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
def test_s3_paths():
    ignore_this_file = temp_filename()
    with open(ignore_this_file, 'w') as f:
        f.write("Not really")

    # Simple case - no sub dir, default filename pattern
    with new_test_dataset(2) as test_ds:
        s3path = test_ds.copy_to_s3()
        job = build_registration_job(s3path, mode=DatasetValidationMode.SAMPLE)
        job.prerun()
        assert job.parts == test_ds.expected_parts
        assert sorted([sp.path for sp in job.sampled_parts]) == \
               sorted([f"{s3path}{fname}" for fname in test_ds.basename_files])

        # Add a file which shouldn't be picked up
        test_ds.bucket.upload_file(ignore_this_file, 'ignore.pq')
        job = build_registration_job(s3path, mode=DatasetValidationMode.SAMPLE)
        job.prerun()
        assert job.parts == test_ds.expected_parts

    # Now with non-default patterns
    with new_test_dataset(8, prefix="blah-", suffix=".pq") as test_ds:
        s3path = test_ds.copy_to_s3()
        # Add files to be ignored
        test_ds.bucket.upload_file(ignore_this_file, 'ignore.parquet')
        test_ds.bucket.upload_file(ignore_this_file, 'sub1/ignore.pq')
        test_ds.bucket.upload_file(ignore_this_file, 'sub2/ignore.parquet')
        test_ds.bucket.upload_file(ignore_this_file, 'sub1/sub11/ignore.pq')

        # Lookup in bucket root
        job = build_registration_job(s3path, pattern="blah*pq", mode=DatasetValidationMode.SAMPLE)
        job.prerun()
        assert job.parts == test_ds.expected_parts

        # Lookup in nested keys
        test_ds.copy_to_s3('sub1')
        test_ds.copy_to_s3('sub1/sub11')
        job = build_registration_job(s3path + 'sub1', pattern="b*.pq", mode=DatasetValidationMode.SAMPLE)
        job.prerun()
        assert job.parts == test_ds.expected_parts

        job = build_registration_job(s3path + 'sub1/sub11/', pattern="b*.pq", mode=DatasetValidationMode.SAMPLE)
        job.prerun()
        assert job.parts == test_ds.expected_parts

        # Ensure registration can actually run on the found paths
        with registration_job_invoker(job) as invoker:
            job_result = invoker.run_all_stages()  # Asserts success by default

        # Key (dir) exists but no matching files
        job = build_registration_job(s3path + 'sub2/', pattern="b*.pq", mode=DatasetValidationMode.SAMPLE)
        job.prerun()
        assert not job.parts


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
@pytest.mark.xfail(strict=True)
def test_s3_missing_bucket_failure():
    job = build_registration_job('s3://whatsthat/', mode=DatasetValidationMode.SAMPLE)
    job.prerun()


@pytest.mark.skipif(SKIP_S3_TESTS, reason="Skipping mock S3 tests")
@pytest.mark.xfail(strict=True)
def test_s3_missing_key_failure():
    with new_test_dataset(2) as test_ds:
        s3path = test_ds.copy_to_s3('exists')
        job = build_registration_job(s3path + '/notreally', mode=DatasetValidationMode.SAMPLE)
        job.prerun()



"""
Execute a single query task.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import time
from typing import List, cast, Optional
from pandas import DataFrame
from frocket.common.dataset import DatasetPartId
from frocket.common.metrics import MetricName, PartSelectMethodLabel
from frocket.common.tasks.base import TaskStatus, TaskAttemptId, BaseTaskRequest
from frocket.common.tasks.query import PartSelectionMode, QueryTaskRequest, QueryResult, QueryTaskResult
from frocket.engine.query_engine import QueryEngine
from frocket.worker.runners.base_task_runner import BaseTaskRunner, TaskRunnerContext
from frocket.worker.runners.part_loader import FilterPredicate

logger = logging.getLogger(__name__)


class QueryTaskRunner(BaseTaskRunner):
    def __init__(self, req: BaseTaskRequest, ctx: TaskRunnerContext):
        super().__init__(req, ctx)
        self._req = cast(QueryTaskRequest, req)  # Avoid type warnings
        self._dataset_part_id: Optional[DatasetPartId] = None
        self._query_result: Optional[QueryResult] = None

    def _do_run(self):
        self._set_part_to_load()
        self._update_status(TaskStatus.LOADING_DATA)
        with self._ctx.metrics.measure(MetricName.TASK_TOTAL_LOAD_SECONDS):
            df = self._load(needed_columns=self._req.used_columns,
                            load_as_categoricals=self._req.load_as_categoricals)

        self._update_status(TaskStatus.RUNNING_QUERY)
        with self._ctx.metrics.measure(MetricName.TASK_RUN_QUERY_SECONDS):
            engine = QueryEngine(self._req.dataset.group_id_column, self._req.dataset.timestamp_column)
            engine_result = engine.run(df, self._req.query)
        self._query_result = engine_result

    def _set_part_to_load(self) -> None:
        task_attempt_no = self._req.attempt_no
        if self._req.mode == PartSelectionMode.SET_BY_INVOKER:
            part_id = self._req.invoker_set_part
            actual_select_method = PartSelectMethodLabel.SET_BY_INVOKER
        elif self._req.mode == PartSelectionMode.SELECTED_BY_WORKER:
            actual_select_method, part_id = self._select_part_myself()
            logger.info(f"Worker selected part: method: {actual_select_method}, file ID: {part_id}, "
                        f"task attempt no.: {task_attempt_no}")
        else:
            raise Exception(f"Don't know how to handle request mode: {self._req.mode}")

        if not part_id:
            raise Exception("No part to load")

        self._ctx.metrics.set_label_enum(actual_select_method)
        self._dataset_part_id = part_id
        self._task_attempt_id = TaskAttemptId(part_id.part_idx, task_attempt_no)

    def _select_part_myself(self):
        """See configuration guide for 'preflight' concept. In general, that's a configurable time period in self-select
        part mode, where 'warm' workers can select the candidates they wish without interruption."""
        time_left_in_preflight = self._ctx.preflight_duration_seconds - BaseTaskRunner.time_since_invocation(self._req)
        candidates = self._ctx.part_loader.get_cached_candidates(self._req.dataset.id)
        sleep_time = 0
        if not candidates and time_left_in_preflight > 0:
            logger.info("Got no candidates but we're still during preflight"
                        f", so sleeping for {time_left_in_preflight} seconds")
            sleep_time = time_left_in_preflight

        if sleep_time:
            time.sleep(time_left_in_preflight)
        self._ctx.metrics.set_metric(MetricName.TASK_PREFLIGHT_SLEEP_SECONDS, sleep_time)

        # If a worker got some candidates, we still gonna try to grab them even if preflight time has ended
        selected_part = self._ctx.datastore.self_select_part(self._req.request_id, self._req.attempt_no, candidates)
        if not selected_part.part_id:
            # Not supposed to happen, unless there's a retry mechanism gone awry
            raise Exception("Got no part for me!")

        if candidates:
            if not selected_part.random:
                actual_select_method = PartSelectMethodLabel.SPECIFIC_CANDIDATE
            else:
                actual_select_method = PartSelectMethodLabel.RANDOM_CANDIDATES_TAKEN
        else:
            actual_select_method = PartSelectMethodLabel.RANDOM_NO_CANDIDATES

        return actual_select_method, selected_part.part_id

    def _load(self, needed_columns: List[str] = None, load_as_categoricals: List[str] = None) -> DataFrame:
        filters = self._predicate_pushdown_filters()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Filters used when loading: {filters}")
            logger.debug(f"Columns to explicitly load as categorical: {load_as_categoricals}")

        df = self._ctx.part_loader.load_dataframe(file_id=self._dataset_part_id, metrics=self._ctx.metrics,
                                                  needed_columns=needed_columns, filters=filters,
                                                  load_as_categoricals=load_as_categoricals)
        self._ctx.metrics.set_metric(MetricName.SCANNED_ROWS, len(df))
        self._ctx.metrics.set_metric(MetricName.SCANNED_GROUPS, df[self._req.dataset.group_id_column].nunique())
        return df

    def _predicate_pushdown_filters(self):
        """
        Build PyArrow-compatible pushdown predicates to pass the part loader.
        An important reminder here is that any filter applied would affect not just conditions/sequences, but also
        any defined aggregations - meaning it's suitable for limiting scope to the (optional) query timeframe,
        but should be evaluated carefully for any other optimizations.
        """
        filters = []
        timeframe = self._req.query.get('timeframe', None)
        if timeframe:
            fromtime = timeframe.get('from', None)
            if fromtime is not None:
                filters.append(FilterPredicate(column=self._req.dataset.timestamp_column, op='>=', value=fromtime))
            totime = timeframe.get('to', None)
            if totime is not None:
                filters.append(FilterPredicate(column=self._req.dataset.timestamp_column, op='<', value=totime))

        return filters if len(filters) > 0 else None

    def _build_result(self, base_attributes):
        return QueryTaskResult(
            **base_attributes,
            query_result=self._query_result)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import abstractmethod, ABCMeta
from dataclasses import dataclass
from typing import List, Dict, Set, Optional, Union
from frocket.common.tasks.base import TaskStatus, BaseTaskResult, TaskAttemptId, TaskStatusUpdate, BaseTaskRequest
from frocket.common.dataset import DatasetInfo, DatasetPartsInfo, DatasetPartId, DatasetShortSchema, DatasetSchema
from frocket.common.serializable import SerializableDataClass

DEFAULT_QUEUE = 'default'
DEFAULT_DEQUEUE_WAIT_TIME = 60


# Used in PartSelectionMode.SELECTED_BY_WORKER
@dataclass(frozen=True)
class WorkerSelectedPart(SerializableDataClass):
    part_id: DatasetPartId
    random: bool
    task_attempt_no: int


class Datastore(metaclass=ABCMeta):
    """
    Interface to the data store, which holds:

    * The list, metadata and schema of all registered datasets
    * For running jobs:
      * Task statuses and results
      * Atomic attempt counter for retried tasks
      * For jobs running in mode PartSelectionMode.SELECTED_BY_WORKER, the manifest of available tasks to select from.
      * When the system is configured to use the 'work_queue' invoker (rather than 'aws_lambda'), the datastore also
        provides the queue through which tasks are enqueued by the invoker and picked up by the workers, like a very
        simplistic queue management system.

    The datastore is not for storing the actual dataset or other persistent large data.
    """
    @abstractmethod
    def write_dataset_info(self, dataset: DatasetInfo, parts: DatasetPartsInfo, schema: DatasetSchema) -> None:
        pass

    @abstractmethod
    def remove_dataset_info(self, name: str) -> bool:
        pass

    @abstractmethod
    def dataset_info(self, name: str) -> DatasetInfo:
        pass

    @abstractmethod
    def dataset_parts_info(self, ds: DatasetInfo) -> DatasetPartsInfo:
        pass

    @abstractmethod
    def schema(self, ds: DatasetInfo) -> DatasetSchema:
        pass

    @abstractmethod
    def short_schema(self, ds: DatasetInfo) -> DatasetShortSchema:
        pass

    @abstractmethod
    def last_used(self, ds: DatasetInfo) -> int:
        pass

    @abstractmethod
    def mark_used(self, ds: DatasetInfo):
        pass

    @abstractmethod
    def datasets(self) -> List[DatasetInfo]:
        pass

    @abstractmethod
    def enqueue(self, requests: List[BaseTaskRequest], queue: str = DEFAULT_QUEUE) -> None:
        pass

    @abstractmethod
    def dequeue(self, queue: str = DEFAULT_QUEUE, timeout: int = DEFAULT_DEQUEUE_WAIT_TIME) -> BaseTaskRequest:
        pass

    @abstractmethod
    def update_task_status(self, reqid: str,
                           tasks: Union[TaskAttemptId, List[TaskAttemptId]], status: TaskStatus) -> None:
        pass

    @abstractmethod
    def tasks_status(self, reqid: str) -> Dict[TaskAttemptId, TaskStatusUpdate]:
        pass

    @abstractmethod
    def write_task_result(self, reqid: str, taskid: TaskAttemptId, result: BaseTaskResult) -> None:
        pass

    @abstractmethod
    def task_results(self, reqid: str) -> Dict[TaskAttemptId, BaseTaskResult]:
        pass

    @abstractmethod
    def increment_attempt(self, reqid: str, part_idx: int) -> int:
        pass

    @abstractmethod
    def publish_for_worker_selection(self, reqid: str, attempt_round: int, parts: Set[DatasetPartId]) -> None:
        pass

    @abstractmethod
    def self_select_part(self, reqid: str, attempt_round: int,
                         candidates: Set[DatasetPartId] = None) -> Optional[WorkerSelectedPart]:
        pass

    @abstractmethod
    def cleanup_request_data(self, reqid: str) -> None:
        pass
